#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/03/2004
#
#------------------------------------------------------------------------------

""" Defines the tree node descriptor used by the tree editor and tree editor
    factory classes.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    AdaptedTo,
    Adapter,
    Any,
    Bool,
    Callable,
    Either,
    HasPrivateTraits,
    Instance,
    Interface,
    isinterface,
    List,
    Property,
    Str,
    cached_property)

from traits.trait_base import SequenceTypes, get_resource_path, xgetattr, xsetattr

from .ui_traits import AView

#-------------------------------------------------------------------------
#  'TreeNode' class:
#-------------------------------------------------------------------------


class TreeNode(HasPrivateTraits):
    """ Represents a tree node. Used by the tree editor and tree editor factory
        classes.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Name of trait containing children (if '', the node is a leaf).
    children = Str

    # Either the name of a trait containing a label, or a constant label, if
    # the string starts with '='.
    label = Str

    # The name of a trait containing a list of labels for any columns.
    column_labels = Str

    # Either the name of a trait containing a tooltip, or constant tooltip, if
    # the string starts with '='.
    tooltip = Str

    # Name to use for a new instance
    name = Str

    # Can the object's children be renamed?
    rename = Bool(True)

    # Can the object be renamed?
    rename_me = Bool(True)

    # Can the object's children be copied?
    copy = Bool(True)

    # Can the object's children be deleted?
    delete = Bool(True)

    # Can the object be deleted (if its parent allows it)?
    delete_me = Bool(True)

    # Can children be inserted (vs. appended)?
    insert = Bool(True)

    # Should tree nodes be automatically opened (expanded)?
    auto_open = Bool(False)

    # Automatically close sibling tree nodes?
    auto_close = Bool(False)

    # List of object classes than can be added or copied
    add = List(Any)

    # List of object classes that can be moved
    move = List(Any)

    # List of object classes and/or interfaces that the node applies to
    node_for = List(Any)

    # Tuple of object classes that the node applies to
    node_for_class = Property(depends_on='node_for')

    # List of object interfaces that the node applies to
    node_for_interface = Property(depends_on='node_for')

    # Function for formatting the label
    formatter = Callable

    # Functions for formatting the other columns.
    column_formatters = List(Either(None, Callable))

    # Function for formatting the tooltip
    tooltip_formatter = Callable

    # Function for handling selecting an object
    on_select = Callable

    # Function for handling clicking an object
    on_click = Callable

    # Function for handling double-clicking an object
    on_dclick = Callable

    # Function for handling activation of an object
    # (double-click or Enter key press when node is in focus)
    on_activated = Callable

    # View to use for editing the object
    view = AView

    # Right-click context menu. The value can be one of:
    #
    # - Instance( Menu ): Use this menu as the context menu
    # - None: Use the default context menu
    # - False: Do not display a context menu
    menu = Any

    # Name of leaf item icon
    icon_item = Str('<item>')

    # Name of group item icon
    icon_group = Str('<group>')

    # Name of opened group item icon
    icon_open = Str('<open>')

    # Resource path used to locate the node icon
    icon_path = Str

    # Selector or name for background color
    background = Any

    # Selector or name for foreground color
    foreground = Any

    # fixme: The 'menu' trait should really be defined as:
    #        Instance( 'traitsui.menu.MenuBar' ), but it doesn't work
    #        right currently.

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, **traits):
        super(TreeNode, self).__init__(**traits)
        if self.icon_path == '':
            self.icon_path = get_resource_path()

    #-- Property Implementations ---------------------------------------------

    @cached_property
    def _get_node_for_class(self):
        return tuple([klass for klass in self.node_for
                      if not isinterface(klass)])

    @cached_property
    def _get_node_for_interface(self):
        return [klass for klass in self.node_for
                if isinterface(klass)]

    #-- Overridable Methods: -------------------------------------------------

    #-------------------------------------------------------------------------
    #  Returns whether chidren of this object are allowed or not:
    #-------------------------------------------------------------------------

    def allows_children(self, object):
        """ Returns whether this object can have children.
        """
        return (self.children != '')

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self, object):
        """ Returns whether the object has children.
        """
        return (len(self.get_children(object)) > 0)

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def get_children(self, object):
        """ Gets the object's children.
        """
        return getattr(object, self.children)

    #-------------------------------------------------------------------------
    #  Gets the object's children identifier:
    #-------------------------------------------------------------------------

    def get_children_id(self, object):
        """ Gets the object's children identifier.
        """
        return self.children

    #-------------------------------------------------------------------------
    #  Appends a child to the object's children:
    #-------------------------------------------------------------------------

    def append_child(self, object, child):
        """ Appends a child to the object's children.
        """
        getattr(object, self.children).append(child)

    #-------------------------------------------------------------------------
    #  Inserts a child into the object's children:
    #-------------------------------------------------------------------------

    def insert_child(self, object, index, child):
        """ Inserts a child into the object's children.
        """
        getattr(object, self.children)[index: index] = [child]

    #-------------------------------------------------------------------------
    #  Confirms that a specified object can be deleted or not:
    #  Result = True:  Delete object with no further prompting
    #         = False: Do not delete object
    #         = other: Take default action (may prompt user to confirm delete)
    #-------------------------------------------------------------------------

    def confirm_delete(self, object):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """
        return None

    #-------------------------------------------------------------------------
    #  Deletes a child at a specified index from the object's children:
    #-------------------------------------------------------------------------

    def delete_child(self, object, index):
        """ Deletes a child at a specified index from the object's children.
        """
        del getattr(object, self.children)[index]

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children replaced' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_replaced(self, object, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
        specified object.
        """
        object.on_trait_change(listener, self.children, remove=remove,
                               dispatch='fast_ui')

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children changed' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_changed(self, object, listener, remove):
        """ Sets up or removes a listener for children being changed on a
        specified object.
        """
        object.on_trait_change(listener, self.children + '_items',
                               remove=remove, dispatch='fast_ui')

    #-------------------------------------------------------------------------
    #  Gets the label to display for a specified object:
    #-------------------------------------------------------------------------

    def get_label(self, object):
        """ Gets the label to display for a specified object.
        """
        label = self.label
        if label[:1] == '=':
            return label[1:]

        label = xgetattr(object, label, '')

        if self.formatter is None:
            return label

        return self.formatter(object, label)

    #-------------------------------------------------------------------------
    #  Sets the label for a specified object:
    #-------------------------------------------------------------------------

    def set_label(self, object, label):
        """ Sets the label for a specified object.
        """
        label_name = self.label
        if label_name[:1] != '=':
            xsetattr(object, label_name, label)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'label changed' on a specified object:
    #-------------------------------------------------------------------------

    def when_label_changed(self, object, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
        specified object.
        """
        label = self.label
        if label[:1] != '=':
            object.on_trait_change(listener, label, remove=remove,
                                   dispatch='ui')

    def get_column_labels(self, object):
        """ Get the labels for any columns that have been defined.
        """
        trait = self.column_labels
        labels = xgetattr(object, trait, [])
        formatted = []
        for formatter, label in map(None, self.column_formatters, labels):
            # If the list of column formatters is shorter than the list of
            # labels, then map(None) will extend it with Nones. Just pass the label
            # as preformatted. Similarly, explicitly using None in the list will
            # pass through the item.
            if formatter is None:
                formatted.append(label)
            else:
                formatted.append(formatter(label))
        return formatted

    def when_column_labels_change(self, object, listener, remove):
        """ Sets up or removes a listener for the column labels being changed on
        a specified object.

        This will fire when either the list is reassigned or when it is
        modified. I.e., it listens both to the trait change event and the
        trait_items change event. Implement the listener appropriately to handle
        either case.
        """
        trait = self.column_labels
        if trait != '':
            object.on_trait_change(
                listener, trait, remove=remove, dispatch='ui')
            object.on_trait_change(
                listener,
                trait + '_items',
                remove=remove,
                dispatch='ui')

    #-------------------------------------------------------------------------
    #  Gets the tooltip to display for a specified object:
    #-------------------------------------------------------------------------

    def get_tooltip(self, object):
        """ Gets the tooltip to display for a specified object.
        """
        tooltip = self.tooltip
        if tooltip == '':
            return tooltip

        if tooltip[:1] == '=':
            return tooltip[1:]

        tooltip = xgetattr(object, tooltip, '')

        if self.tooltip_formatter is None:
            return tooltip

        return self.tooltip_formatter(object, tooltip)

    #-------------------------------------------------------------------------
    #  Returns the icon for a specified object:
    #-------------------------------------------------------------------------

    def get_icon(self, object, is_expanded):
        """ Returns the icon for a specified object.
        """
        if not self.allows_children(object):
            return self.icon_item

        if is_expanded:
            return self.icon_open

        return self.icon_group

    #-------------------------------------------------------------------------
    #  Returns the path used to locate an object's icon:
    #-------------------------------------------------------------------------

    def get_icon_path(self, object):
        """ Returns the path used to locate an object's icon.
        """
        return self.icon_path

    #-------------------------------------------------------------------------
    #  Returns the name to use when adding a new object instance (displayed in
    #  the 'New' submenu):
    #-------------------------------------------------------------------------

    def get_name(self, object):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return self.name

    #-------------------------------------------------------------------------
    #  Gets the View to use when editing an object:
    #-------------------------------------------------------------------------

    def get_view(self, object):
        """ Gets the view to use when editing an object.
        """
        return self.view

    #-------------------------------------------------------------------------
    #  Returns the right-click context menu for an object:
    #-------------------------------------------------------------------------

    def get_menu(self, object):
        """ Returns the right-click context menu for an object.
        """
        return self.menu

    def get_background(self, object):
        background = self.background
        if isinstance(background, basestring):
            background = getattr(object, background, background)
        return background

    def get_foreground(self, object):
        foreground = self.foreground
        if isinstance(foreground, basestring):
            foreground = getattr(object, foreground, foreground)
        return foreground

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be renamed:
    #-------------------------------------------------------------------------

    def can_rename(self, object):
        """ Returns whether the object's children can be renamed.
        """
        return self.rename

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be renamed:
    #-------------------------------------------------------------------------

    def can_rename_me(self, object):
        """ Returns whether the object can be renamed.
        """
        return self.rename_me

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be copied:
    #-------------------------------------------------------------------------

    def can_copy(self, object):
        """ Returns whether the object's children can be copied.
        """
        return self.copy

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be deleted:
    #-------------------------------------------------------------------------

    def can_delete(self, object):
        """ Returns whether the object's children can be deleted.
        """
        return self.delete

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be deleted:
    #-------------------------------------------------------------------------

    def can_delete_me(self, object):
        """ Returns whether the object can be deleted.
        """
        return self.delete_me

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be inserted (or just
    #  appended):
    #-------------------------------------------------------------------------

    def can_insert(self, object):
        """ Returns whether the object's children can be inserted (vs.
        appended).
        """
        return self.insert

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-opened:
    #-------------------------------------------------------------------------

    def can_auto_open(self, object):
        """ Returns whether the object's children should be automatically
        opened.
        """
        return self.auto_open

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-closed:
    #-------------------------------------------------------------------------

    def can_auto_close(self, object):
        """ Returns whether the object's children should be automatically
        closed.
        """
        return self.auto_close

    #-------------------------------------------------------------------------
    #  Returns whether or not this is the node that should handle a specified
    #  object:
    #-------------------------------------------------------------------------

    def is_node_for(self, object):
        """ Returns whether this is the node that handles a specified object.
        """
        return (isinstance(object, self.node_for_class) or
                object.has_traits_interface(*self.node_for_interface))

    #-------------------------------------------------------------------------
    #  Returns whether a given 'add_object' can be added to an object:
    #-------------------------------------------------------------------------

    def can_add(self, object, add_object):
        """ Returns whether a given object is droppable on the node.
        """
        klass = self._class_for(add_object)
        if self.is_addable(klass):
            return True

        for item in self.move:
            if type(item) in SequenceTypes:
                item = item[0]
            if issubclass(klass, item):
                return True

        return False

    #-------------------------------------------------------------------------
    #  Returns the list of classes that can be added to the object:
    #-------------------------------------------------------------------------

    def get_add(self, object):
        """ Returns the list of classes that can be added to the object.
        """
        return self.add

    #-------------------------------------------------------------------------
    #  Returns the 'draggable' version of a specified object:
    #-------------------------------------------------------------------------

    def get_drag_object(self, object):
        """ Returns a draggable version of a specified object.
        """
        return object

    #-------------------------------------------------------------------------
    #  Returns a droppable version of a specified object:
    #-------------------------------------------------------------------------

    def drop_object(self, object, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        klass = self._class_for(dropped_object)
        if self.is_addable(klass):
            return dropped_object

        for item in self.move:
            if type(item) in SequenceTypes:
                if issubclass(klass, item[0]):
                    return item[1](object, dropped_object)
            elif issubclass(klass, item):
                return dropped_object

        return dropped_object

    #-------------------------------------------------------------------------
    #  Handles an object being selected:
    #-------------------------------------------------------------------------

    def select(self, object):
        """ Handles an object being selected.
        """
        if self.on_select is not None:
            self.on_select(object)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being clicked:
    #-------------------------------------------------------------------------

    def click(self, object):
        """ Handles an object being clicked.
        """
        if self.on_click is not None:
            self.on_click(object)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being double-clicked:
    #-------------------------------------------------------------------------

    def dclick(self, object):
        """ Handles an object being double-clicked.
        """
        if self.on_dclick is not None:
            self.on_dclick(object)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being activated:
    #-------------------------------------------------------------------------

    def activated(self, object):
        """ Handles an object being activated.
        """
        if self.on_activated is not None:
            self.on_activated(object)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Returns whether or not a specified object class can be added to the node:
    #-------------------------------------------------------------------------

    def is_addable(self, klass):
        """ Returns whether a specified object class can be added to the node.
        """
        for item in self.add:
            if type(item) in SequenceTypes:
                item = item[0]

            if issubclass(klass, item):
                return True

        return False

    #-------------------------------------------------------------------------
    #  Returns the class of an object:
    #-------------------------------------------------------------------------

    def _class_for(self, object):
        """ Returns the class of an object.
        """
        if isinstance(object, type):
            return object

        return object.__class__

#-------------------------------------------------------------------------
#  'ITreeNode' class
#-------------------------------------------------------------------------


class ITreeNode(Interface):

    def allows_children(self):
        """ Returns whether this object can have children.
        """

    def has_children(self):
        """ Returns whether the object has children.
        """

    def get_children(self):
        """ Gets the object's children.
        """

    def get_children_id(self):
        """ Gets the object's children identifier.
        """

    def append_child(self, child):
        """ Appends a child to the object's children.
        """

    def insert_child(self, index, child):
        """ Inserts a child into the object's children.
        """

    def confirm_delete(self):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """

    def delete_child(self, index):
        """ Deletes a child at a specified index from the object's children.
        """

    def when_children_replaced(self, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
            specified object.
        """

    def when_children_changed(self, listener, remove):
        """ Sets up or removes a listener for children being changed on a
            specified object.
        """

    def get_label(self):
        """ Gets the label to display for a specified object.
        """

    def set_label(self, label):
        """ Sets the label for a specified object.
        """

    def when_label_changed(self, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
            specified object.
        """

    def get_column_labels(self, object):
        """ Get the labels for any columns that have been defined.
        """

    def when_column_labels_change(self, object, listener, remove):
        """ Sets up or removes a listener for the column labels being changed on
        a specified object.

        This will fire when either the list is reassigned or when it is
        modified. I.e., it listens both to the trait change event and the
        trait_items change event. Implement the listener appropriately to handle
        either case.
        """

    def get_tooltip(self):
        """ Gets the tooltip to display for a specified object.
        """

    def get_icon(self, is_expanded):
        """ Returns the icon for a specified object.
        """

    def get_icon_path(self):
        """ Returns the path used to locate an object's icon.
        """

    def get_name(self):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """

    def get_view(self):
        """ Gets the view to use when editing an object.
        """

    def get_menu(self):
        """ Returns the right-click context menu for an object.
        """

    def can_rename(self):
        """ Returns whether the object's children can be renamed.
        """

    def can_rename_me(self):
        """ Returns whether the object can be renamed.
        """

    def can_copy(self):
        """ Returns whether the object's children can be copied.
        """

    def can_delete(self):
        """ Returns whether the object's children can be deleted.
        """

    def can_delete_me(self):
        """ Returns whether the object can be deleted.
        """

    def can_insert(self):
        """ Returns whether the object's children can be inserted (vs.
            appended).
        """

    def can_auto_open(self):
        """ Returns whether the object's children should be automatically
            opened.
        """

    def can_auto_close(self):
        """ Returns whether the object's children should be automatically
            closed.
        """

    def can_add(self, add_object):
        """ Returns whether a given object is droppable on the node.
        """

    def get_add(self):
        """ Returns the list of classes that can be added to the object.
        """

    def get_drag_object(self):
        """ Returns a draggable version of a specified object.
        """

    def drop_object(self, dropped_object):
        """ Returns a droppable version of a specified object.
        """

    def select(self):
        """ Handles an object being selected.
        """

    def click(self):
        """ Handles an object being clicked.
        """

    def dclick(self):
        """ Handles an object being double-clicked.
        """

    def activated(self):
        """ Handles an object being activated.
        """

#-------------------------------------------------------------------------
#  'ITreeNodeAdapter' class
#-------------------------------------------------------------------------


class ITreeNodeAdapter(Adapter):
    """ Abstract base class for an adapter that implements the ITreeNode
        interface.

        Usage:

        1. Create a subclass of ITreeNodeAdapter.
        2. Add an 'adapts( xxx_class, ITreeNode )' declaration (usually placed
           right after the 'class' statement) to define what class (or classes)
           this is an ITreeNode adapter for.
        3. Override any of the following methods as necessary, using the
           'self.adaptee' trait to access the adapted object if needed.

       Note: This base class implements all of the ITreeNode interface methods,
       but does not necessarily provide useful implementations for all of the
       methods. It allows you to get a new adapter class up and running quickly,
       but you should carefully review your final adapter implementation class
       to make sure it behaves correctly in your application.
    """

    def allows_children(self):
        """ Returns whether this object can have children.
        """
        return False

    def has_children(self):
        """ Returns whether the object has children.
        """
        return False

    def get_children(self):
        """ Gets the object's children.
        """
        return []

    def get_children_id(self):
        """ Gets the object's children identifier.
        """
        return ''

    def append_child(self, child):
        """ Appends a child to the object's children.
        """
        pass

    def insert_child(self, index, child):
        """ Inserts a child into the object's children.
        """
        pass

    def confirm_delete(self):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """
        return False

    def delete_child(self, index):
        """ Deletes a child at a specified index from the object's children.
        """
        pass

    def when_children_replaced(self, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
            specified object.
        """
        pass

    def when_children_changed(self, listener, remove):
        """ Sets up or removes a listener for children being changed on a
            specified object.
        """
        pass

    def get_label(self):
        """ Gets the label to display for a specified object.
        """
        return 'No label specified'

    def set_label(self, label):
        """ Sets the label for a specified object.
        """
        pass

    def when_label_changed(self, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
            specified object.
        """
        pass

    def get_column_labels(self):
        """ Get the labels for any columns that have been defined.
        """
        return []

    def when_column_labels_change(self, listener, remove):
        """ Sets up or removes a listener for the column labels being changed on
        a specified object.

        This will fire when either the list is reassigned or when it is
        modified. I.e., it listens both to the trait change event and the
        trait_items change event. Implement the listener appropriately to handle
        either case.
        """
        pass

    def get_tooltip(self):
        """ Gets the tooltip to display for a specified object.
        """
        return ''

    def get_icon(self, is_expanded):
        """ Returns the icon for a specified object.
        """
        return '<item>'

    def get_icon_path(self):
        """ Returns the path used to locate an object's icon.
        """
        return ''

    def get_name(self):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return ''

    def get_view(self):
        """ Gets the view to use when editing an object.
        """
        return None

    def get_menu(self):
        """ Returns the right-click context menu for an object.
        """
        return None

    def get_background(self):
        """ Returns the background for object
        """
        return None

    def get_foreground(self):
        """ Returns the foreground for object
        """
        return None

    def can_rename(self):
        """ Returns whether the object's children can be renamed.
        """
        return False

    def can_rename_me(self):
        """ Returns whether the object can be renamed.
        """
        return False

    def can_copy(self):
        """ Returns whether the object's children can be copied.
        """
        return False

    def can_delete(self):
        """ Returns whether the object's children can be deleted.
        """
        return False

    def can_delete_me(self):
        """ Returns whether the object can be deleted.
        """
        return False

    def can_insert(self):
        """ Returns whether the object's children can be inserted (vs.
            appended).
        """
        return False

    def can_auto_open(self):
        """ Returns whether the object's children should be automatically
            opened.
        """
        return False

    def can_auto_close(self):
        """ Returns whether the object's children should be automatically
            closed.
        """
        return False

    def can_add(self, add_object):
        """ Returns whether a given object is droppable on the node.
        """
        return False

    def get_add(self):
        """ Returns the list of classes that can be added to the object.
        """
        return []

    def get_drag_object(self):
        """ Returns a draggable version of a specified object.
        """
        return self.adaptee

    def drop_object(self, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        return dropped_object

    def select(self):
        """ Handles an object being selected.
        """
        pass

    def click(self):
        """ Handles an object being clicked.
        """
        pass

    def dclick(self):
        """ Handles an object being double-clicked.
        """
        pass

    def activated(self):
        """ Handles an object being activated.
        """
        pass

#-------------------------------------------------------------------------
#  'ITreeNodeAdapterBridge' class
#-------------------------------------------------------------------------


class ITreeNodeAdapterBridge(HasPrivateTraits):
    """ Private class for use by a toolkit-specific implementation of the
        TreeEditor to allow bridging the TreeNode interface used by the editor
        to the ITreeNode interface used by object adapters.
    """

    # The ITreeNode adapter being bridged:
    adapter = AdaptedTo(ITreeNode)

    #-- TreeNode implementation ----------------------------------------------

    def allows_children(self, object):
        """ Returns whether this object can have children.
        """
        return self.adapter.allows_children()

    def has_children(self, object):
        """ Returns whether the object has children.
        """
        return self.adapter.has_children()

    def get_children(self, object):
        """ Gets the object's children.
        """
        return self.adapter.get_children()

    def get_children_id(self, object):
        """ Gets the object's children identifier.
        """
        return self.adapter.get_children_id()

    def append_child(self, object, child):
        """ Appends a child to the object's children.
        """
        return self.adapter.append_child(child)

    def insert_child(self, object, index, child):
        """ Inserts a child into the object's children.
        """
        return self.adapter.insert_child(index, child)

    def confirm_delete(self, object):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """
        return self.adapter.confirm_delete()

    def delete_child(self, object, index):
        """ Deletes a child at a specified index from the object's children.
        """
        return self.adapter.delete_child(index)

    def when_children_replaced(self, object, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
            specified object.
        """
        return self.adapter.when_children_replaced(listener, remove)

    def when_children_changed(self, object, listener, remove):
        """ Sets up or removes a listener for children being changed on a
            specified object.
        """
        return self.adapter.when_children_changed(listener, remove)

    def get_label(self, object):
        """ Gets the label to display for a specified object.
        """
        return self.adapter.get_label()

    def set_label(self, object, label):
        """ Sets the label for a specified object.
        """
        return self.adapter.set_label(label)

    def when_label_changed(self, object, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
            specified object.
        """
        return self.adapter.when_label_changed(listener, remove)

    def get_column_labels(self, object):
        """ Get the labels for any columns that have been defined.
        """
        return self.adapter.get_column_labels()

    def when_column_labels_change(self, object, listener, remove):
        """ Sets up or removes a listener for the column labels being changed on
        a specified object.

        This will fire when either the list is reassigned or when it is
        modified. I.e., it listens both to the trait change event and the
        trait_items change event. Implement the listener appropriately to handle
        either case.
        """
        return self.adapter.when_column_labels_change(listener, remove)

    def get_tooltip(self, object):
        """ Gets the tooltip to display for a specified object.
        """
        return self.adapter.get_tooltip()

    def get_icon(self, object, is_expanded):
        """ Returns the icon for a specified object.
        """
        return self.adapter.get_icon(is_expanded)

    def get_icon_path(self, object):
        """ Returns the path used to locate an object's icon.
        """
        return self.adapter.get_icon_path()

    def get_name(self, object):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return self.adapter.get_name()

    def get_view(self, object):
        """ Gets the view to use when editing an object.
        """
        return self.adapter.get_view()

    def get_menu(self, object):
        """ Returns the right-click context menu for an object.
        """
        return self.adapter.get_menu()

    def get_background(self, object):
        """ Returns the background for object
        """
        return self.adapter.get_background()

    def get_foreground(self, object):
        """ Returns the foreground for object
        """
        return self.adapter.get_foreground()

    def can_rename(self, object):
        """ Returns whether the object's children can be renamed.
        """
        return self.adapter.can_rename()

    def can_rename_me(self, object):
        """ Returns whether the object can be renamed.
        """
        return self.adapter.can_rename_me()

    def can_copy(self, object):
        """ Returns whether the object's children can be copied.
        """
        return self.adapter.can_copy()

    def can_delete(self, object):
        """ Returns whether the object's children can be deleted.
        """
        return self.adapter.can_delete()

    def can_delete_me(self, object):
        """ Returns whether the object can be deleted.
        """
        return self.adapter.can_delete_me()

    def can_insert(self, object):
        """ Returns whether the object's children can be inserted (vs.
            appended).
        """
        return self.adapter.can_insert()

    def can_auto_open(self, object):
        """ Returns whether the object's children should be automatically
            opened.
        """
        return self.adapter.can_auto_open()

    def can_auto_close(self, object):
        """ Returns whether the object's children should be automatically
            closed.
        """
        return self.adapter.can_auto_close()

    def can_add(self, object, add_object):
        """ Returns whether a given object is droppable on the node.
        """
        return self.adapter.can_add(add_object)

    def get_add(self, object):
        """ Returns the list of classes that can be added to the object.
        """
        return self.adapter.get_add()

    def get_drag_object(self, object):
        """ Returns a draggable version of a specified object.
        """
        return self.adapter.get_drag_object()

    def drop_object(self, object, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        return self.adapter.drop_object(dropped_object)

    def select(self, object):
        """ Handles an object being selected.
        """
        return self.adapter.select()

    def click(self, object):
        """ Handles an object being clicked.
        """
        return self.adapter.click()

    def dclick(self, object):
        """ Handles an object being double-clicked.
        """
        return self.adapter.dclick()

    def activated(self, object):
        """ Handles an object being activated.
        """
        return self.adapter.activated()


# FIXME RTK: add the column_labels API to the following TreeNodes, too.


#-------------------------------------------------------------------------
#  'ObjectTreeNode' class
#-------------------------------------------------------------------------

class ObjectTreeNode(TreeNode):

    #-------------------------------------------------------------------------
    #  Returns whether chidren of this object are allowed or not:
    #-------------------------------------------------------------------------

    def allows_children(self, object):
        """ Returns whether this object can have children.
        """
        return object.tno_allows_children(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self, object):
        """ Returns whether the object has children.
        """
        return object.tno_has_children(self)

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def get_children(self, object):
        """ Gets the object's children.
        """
        return object.tno_get_children(self)

    #-------------------------------------------------------------------------
    #  Gets the object's children identifier:
    #-------------------------------------------------------------------------

    def get_children_id(self, object):
        """ Gets the object's children identifier.
        """
        return object.tno_get_children_id(self)

    #-------------------------------------------------------------------------
    #  Appends a child to the object's children:
    #-------------------------------------------------------------------------

    def append_child(self, object, child):
        """ Appends a child to the object's children.
        """
        return object.tno_append_child(self, child)

    #-------------------------------------------------------------------------
    #  Inserts a child into the object's children:
    #-------------------------------------------------------------------------

    def insert_child(self, object, index, child):
        """ Inserts a child into the object's children.
        """
        return object.tno_insert_child(self, index, child)

    #-------------------------------------------------------------------------
    #  Confirms that a specified object can be deleted or not:
    #  Result = True:  Delete object with no further prompting
    #         = False: Do not delete object
    #         = other: Take default action (may prompt user to confirm delete)
    #-------------------------------------------------------------------------

    def confirm_delete(self, object):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """
        return object.tno_confirm_delete(self)

    #-------------------------------------------------------------------------
    #  Deletes a child at a specified index from the object's children:
    #-------------------------------------------------------------------------

    def delete_child(self, object, index):
        """ Deletes a child at a specified index from the object's children.
        """
        return object.tno_delete_child(self, index)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children replaced' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_replaced(self, object, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
            specified object.
        """
        return object.tno_when_children_replaced(self, listener, remove)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children changed' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_changed(self, object, listener, remove):
        """ Sets up or removes a listener for children being changed on a
            specified object.
        """
        return object.tno_when_children_changed(self, listener, remove)

    #-------------------------------------------------------------------------
    #  Gets the label to display for a specified object:
    #-------------------------------------------------------------------------

    def get_label(self, object):
        """ Gets the label to display for a specified object.
        """
        return object.tno_get_label(self)

    #-------------------------------------------------------------------------
    #  Sets the label for a specified object:
    #-------------------------------------------------------------------------

    def set_label(self, object, label):
        """ Sets the label for a specified object.
        """
        return object.tno_set_label(self, label)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'label changed' on a specified object:
    #-------------------------------------------------------------------------

    def when_label_changed(self, object, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
            specified object.
        """
        return object.tno_when_label_changed(self, listener, remove)

    #-------------------------------------------------------------------------
    #  Gets the tooltip to display for a specified object:
    #-------------------------------------------------------------------------

    def get_tooltip(self, object):
        """ Gets the tooltip to display for a specified object.
        """
        return object.tno_get_tooltip(self)

    #-------------------------------------------------------------------------
    #  Returns the icon for a specified object:
    #-------------------------------------------------------------------------

    def get_icon(self, object, is_expanded):
        """ Returns the icon for a specified object.
        """
        return object.tno_get_icon(self, is_expanded)

    #-------------------------------------------------------------------------
    #  Returns the path used to locate an object's icon:
    #-------------------------------------------------------------------------

    def get_icon_path(self, object):
        """ Returns the path used to locate an object's icon.
        """
        return object.tno_get_icon_path(self)

    #-------------------------------------------------------------------------
    #  Returns the name to use when adding a new object instance (displayed in
    #  the 'New' submenu):
    #-------------------------------------------------------------------------

    def get_name(self, object):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return object.tno_get_name(self)

    #-------------------------------------------------------------------------
    #  Gets the View to use when editing an object:
    #-------------------------------------------------------------------------

    def get_view(self, object):
        """ Gets the view to use when editing an object.
        """
        return object.tno_get_view(self)

    #-------------------------------------------------------------------------
    #  Returns the right-click context menu for an object:
    #-------------------------------------------------------------------------

    def get_menu(self, object):
        """ Returns the right-click context menu for an object.
        """
        return object.tno_get_menu(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be renamed:
    #-------------------------------------------------------------------------

    def can_rename(self, object):
        """ Returns whether the object's children can be renamed.
        """
        return object.tno_can_rename(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be renamed:
    #-------------------------------------------------------------------------

    def can_rename_me(self, object):
        """ Returns whether the object can be renamed.
        """
        return object.tno_can_rename_me(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be copied:
    #-------------------------------------------------------------------------

    def can_copy(self, object):
        """ Returns whether the object's children can be copied.
        """
        return object.tno_can_copy(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be deleted:
    #-------------------------------------------------------------------------

    def can_delete(self, object):
        """ Returns whether the object's children can be deleted.
        """
        return object.tno_can_delete(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be deleted:
    #-------------------------------------------------------------------------

    def can_delete_me(self, object):
        """ Returns whether the object can be deleted.
        """
        return object.tno_can_delete_me(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be inserted (or just
    #  appended):
    #-------------------------------------------------------------------------

    def can_insert(self, object):
        """ Returns whether the object's children can be inserted (vs.
        appended).
        """
        return object.tno_can_insert(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-opened:
    #-------------------------------------------------------------------------

    def can_auto_open(self, object):
        """ Returns whether the object's children should be automatically
            opened.
        """
        return object.tno_can_auto_open(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-closed:
    #-------------------------------------------------------------------------

    def can_auto_close(self, object):
        """ Returns whether the object's children should be automatically
            closed.
        """
        return object.tno_can_auto_close(self)

    #-------------------------------------------------------------------------
    #  Returns whether or not this is the node that should handle a specified
    #  object:
    #-------------------------------------------------------------------------

    def is_node_for(self, object):
        """ Returns whether this is the node that should handle a
            specified object.
        """
        if isinstance(object, TreeNodeObject):
            return object.tno_is_node_for(self)

        return False

    #-------------------------------------------------------------------------
    #  Returns whether a given 'add_object' can be added to an object:
    #-------------------------------------------------------------------------

    def can_add(self, object, add_object):
        """ Returns whether a given object is droppable on the node.
        """
        return object.tno_can_add(self, add_object)

    #-------------------------------------------------------------------------
    #  Returns the list of classes that can be added to the object:
    #-------------------------------------------------------------------------

    def get_add(self, object):
        """ Returns the list of classes that can be added to the object.
        """
        return object.tno_get_add(self)

    #-------------------------------------------------------------------------
    #  Returns the 'draggable' version of a specified object:
    #-------------------------------------------------------------------------

    def get_drag_object(self, object):
        """ Returns a draggable version of a specified object.
        """
        return object.tno_get_drag_object(self)

    #-------------------------------------------------------------------------
    #  Returns a droppable version of a specified object:
    #-------------------------------------------------------------------------

    def drop_object(self, object, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        return object.tno_drop_object(self, dropped_object)

    #-------------------------------------------------------------------------
    #  Handles an object being selected:
    #-------------------------------------------------------------------------

    def select(self, object):
        """ Handles an object being selected.
        """
        return object.tno_select(self)

    #-------------------------------------------------------------------------
    #  Handles an object being clicked:
    #-------------------------------------------------------------------------

    def click(self, object):
        """ Handles an object being clicked.
        """
        return object.tno_click(self)

    #-------------------------------------------------------------------------
    #  Handles an object being double-clicked:
    #-------------------------------------------------------------------------

    def dclick(self, object):
        """ Handles an object being double-clicked.
        """
        return object.tno_dclick(self)

    #-------------------------------------------------------------------------
    #  Handles an object being activated:
    #-------------------------------------------------------------------------

    def activated(self, object):
        """ Handles an object being activated.
        """
        return object.tno_activated(self)

#-------------------------------------------------------------------------
#  'TreeNodeObject' class:
#-------------------------------------------------------------------------


class TreeNodeObject(HasPrivateTraits):
    """ Represents the object that corresponds to a tree node.
    """

    #-------------------------------------------------------------------------
    #  Returns whether chidren of this object are allowed or not:
    #-------------------------------------------------------------------------

    def tno_allows_children(self, node):
        """ Returns whether this object allows children.
        """
        return (node.children != '')

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def tno_has_children(self, node):
        """ Returns whether this object has children.
        """
        return (len(self.tno_get_children(node)) > 0)

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def tno_get_children(self, node):
        """ Gets the object's children.
        """
        return getattr(self, node.children)

    #-------------------------------------------------------------------------
    #  Gets the object's children identifier:
    #-------------------------------------------------------------------------

    def tno_get_children_id(self, node):
        """ Gets the object's children identifier.
        """
        return node.children

    #-------------------------------------------------------------------------
    #  Appends a child to the object's children:
    #-------------------------------------------------------------------------

    def tno_append_child(self, node, child):
        """ Appends a child to the object's children.
        """
        getattr(self, node.children).append(child)

    #-------------------------------------------------------------------------
    #  Inserts a child into the object's children:
    #-------------------------------------------------------------------------

    def tno_insert_child(self, node, index, child):
        """ Inserts a child into the object's children.
        """
        getattr(self, node.children)[index: index] = [child]

    #-------------------------------------------------------------------------
    #  Confirms that a specified object can be deleted or not:
    #  Result = True:  Delete object with no further prompting
    #         = False: Do not delete object
    #         = other: Take default action (may prompt user to confirm delete)
    #-------------------------------------------------------------------------

    def tno_confirm_delete(self, node):
        """ Checks whether a specified object can be deleted.

        Returns
        -------
        * **True** if the object should be deleted with no further prompting.
        * **False** if the object should not be deleted.
        * Anything else: Caller should take its default action (which might
          include prompting the user to confirm deletion).
        """
        return None

    #-------------------------------------------------------------------------
    #  Deletes a child at a specified index from the object's children:
    #-------------------------------------------------------------------------

    def tno_delete_child(self, node, index):
        """ Deletes a child at a specified index from the object's children.
        """
        del getattr(self, node.children)[index]

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children replaced' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def tno_when_children_replaced(self, node, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
        specified object.
        """
        self.on_trait_change(listener, node.children, remove=remove,
                             dispatch='fast_ui')

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children changed' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def tno_when_children_changed(self, node, listener, remove):
        """ Sets up or removes a listener for children being changed on a
        specified object.
        """
        self.on_trait_change(listener, node.children + '_items',
                             remove=remove, dispatch='fast_ui')

    #-------------------------------------------------------------------------
    #  Gets the label to display for a specified object:
    #-------------------------------------------------------------------------

    def tno_get_label(self, node):
        """ Gets the label to display for a specified object.
        """
        label = node.label
        if label[:1] == '=':
            return label[1:]

        label = xgetattr(self, label)

        if node.formatter is None:
            return label

        return node.formatter(self, label)

    #-------------------------------------------------------------------------
    #  Sets the label for a specified node:
    #-------------------------------------------------------------------------

    def tno_set_label(self, node, label):
        """ Sets the label for a specified object.
        """
        label_name = node.label
        if label_name[:1] != '=':
            xsetattr(self, label_name, label)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'label changed' on a specified object:
    #-------------------------------------------------------------------------

    def tno_when_label_changed(self, node, listener, remove):
        """ Sets up or removes a listener for  the label being changed on a
        specified object.
        """
        label = node.label
        if label[:1] != '=':
            self.on_trait_change(listener, label, remove=remove,
                                 dispatch='ui')

    #-------------------------------------------------------------------------
    #  Gets the tooltip to display for a specified object:
    #-------------------------------------------------------------------------

    def tno_get_tooltip(self, node):
        """ Gets the tooltip to display for a specified object.
        """
        tooltip = node.tooltip
        if tooltip == '':
            return tooltip

        if tooltip[:1] == '=':
            return tooltip[1:]

        tooltip = xgetattr(self, tooltip)

        if node.tooltip_formatter is None:
            return tooltip

        return node.tooltip_formatter(self, tooltip)

    #-------------------------------------------------------------------------
    #  Returns the icon for a specified object:
    #-------------------------------------------------------------------------

    def tno_get_icon(self, node, is_expanded):
        """ Returns the icon for a specified object.
        """
        if not self.tno_allows_children(node):
            return node.icon_item

        if is_expanded:
            return node.icon_open

        return node.icon_group

    #-------------------------------------------------------------------------
    #  Returns the path used to locate an object's icon:
    #-------------------------------------------------------------------------

    def tno_get_icon_path(self, node):
        """ Returns the path used to locate an object's icon.
        """
        return node.icon_path

    #-------------------------------------------------------------------------
    #  Returns the name to use when adding a new object instance (displayed in
    #  the 'New' submenu):
    #-------------------------------------------------------------------------

    def tno_get_name(self, node):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return node.name

    #-------------------------------------------------------------------------
    #  Gets the View to use when editing an object:
    #-------------------------------------------------------------------------

    def tno_get_view(self, node):
        """ Gets the view to use when editing an object.
        """
        return node.view

    #-------------------------------------------------------------------------
    #  Returns the right-click context menu for an object:
    #-------------------------------------------------------------------------

    def tno_get_menu(self, node):
        """ Returns the right-click context menu for an object.
        """
        return node.menu

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be renamed:
    #-------------------------------------------------------------------------

    def tno_can_rename(self, node):
        """ Returns whether the object's children can be renamed.
        """
        return node.rename

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be renamed:
    #-------------------------------------------------------------------------

    def tno_can_rename_me(self, node):
        """ Returns whether the object can be renamed.
        """
        return node.rename_me

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be copied:
    #-------------------------------------------------------------------------

    def tno_can_copy(self, node):
        """ Returns whether the object's children can be copied.
        """
        return node.copy

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be deleted:
    #-------------------------------------------------------------------------

    def tno_can_delete(self, node):
        """ Returns whether the object's children can be deleted.
        """
        return node.delete

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be deleted:
    #-------------------------------------------------------------------------

    def tno_can_delete_me(self, node):
        """ Returns whether the object can be deleted.
        """
        return node.delete_me

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be inserted (or just
    #  appended):
    #-------------------------------------------------------------------------

    def tno_can_insert(self, node):
        """ Returns whether the object's children can be inserted (vs.
        appended).
        """
        return node.insert

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-opened:
    #-------------------------------------------------------------------------

    def tno_can_auto_open(self, node):
        """ Returns whether the object's children should be automatically
        opened.
        """
        return node.auto_open

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-closed:
    #-------------------------------------------------------------------------

    def tno_can_auto_close(self, node):
        """ Returns whether the object's children should be automatically
        closed.
        """
        return node.auto_close

    #-------------------------------------------------------------------------
    #  Returns whether or not this is the node that should handle a specified
    #  object:
    #-------------------------------------------------------------------------

    def tno_is_node_for(self, node):
        """ Returns whether this is the node that should handle a
            specified object.
        """
        return (isinstance(self, node.node_for_class) or
                self.has_traits_interface(*node.node_for_interface))

    #-------------------------------------------------------------------------
    #  Returns whether a given 'add_object' can be added to an object:
    #-------------------------------------------------------------------------

    def tno_can_add(self, node, add_object):
        """ Returns whether a given object is droppable on the node.
        """
        klass = node._class_for(add_object)
        if node.is_addable(klass):
            return True

        for item in node.move:
            if type(item) in SequenceTypes:
                item = item[0]
            if issubclass(klass, item):
                return True

        return False

    #-------------------------------------------------------------------------
    #  Returns the list of classes that can be added to the object:
    #-------------------------------------------------------------------------

    def tno_get_add(self, node):
        """ Returns the list of classes that can be added to the object.
        """
        return node.add

    #-------------------------------------------------------------------------
    #  Returns the 'draggable' version of a specified object:
    #-------------------------------------------------------------------------

    def tno_get_drag_object(self, node):
        """ Returns a draggable version of a specified object.
        """
        return self

    #-------------------------------------------------------------------------
    #  Returns a droppable version of a specified object:
    #-------------------------------------------------------------------------

    def tno_drop_object(self, node, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        if node.is_addable(dropped_object):
            return dropped_object

        for item in node.move:
            if type(item) in SequenceTypes:
                if isinstance(dropped_object, item[0]):
                    return item[1](self, dropped_object)
            else:
                if isinstance(dropped_object, item):
                    return dropped_object

    #-------------------------------------------------------------------------
    #  Handles an object being selected:
    #-------------------------------------------------------------------------

    def tno_select(self, node):
        """ Handles an object being selected.
        """
        if node.on_select is not None:
            node.on_select(self)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being clicked:
    #-------------------------------------------------------------------------

    def tno_click(self, node):
        """ Handles an object being clicked.
        """
        if node.on_click is not None:
            node.on_click(self)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being double-clicked:
    #-------------------------------------------------------------------------

    def tno_dclick(self, node):
        """ Handles an object being double-clicked.
        """
        if node.on_dclick is not None:
            node.on_dclick(self)
            return None

        return True

    #-------------------------------------------------------------------------
    #  Handles an object being activated:
    #-------------------------------------------------------------------------

    def tno_activated(self, node):
        """ Handles an object being activated.
        """
        if node.on_activated is not None:
            node.on_activated(self)
            return None

        return True

#-------------------------------------------------------------------------
#  'MultiTreeNode' object:
#-------------------------------------------------------------------------


class MultiTreeNode(TreeNode):

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # TreeNode that applies to the base object itself
    root_node = Instance(TreeNode)

    # List of TreeNodes (one for each sub-item list)
    nodes = List(TreeNode)

    #-------------------------------------------------------------------------
    #  Returns whether chidren of this object are allowed or not:
    #-------------------------------------------------------------------------

    def allows_children(self, object):
        """ Returns whether this object can have children (True for this
        class).
        """
        return True

    #-------------------------------------------------------------------------
    #  Returns whether or not the object has children:
    #-------------------------------------------------------------------------

    def has_children(self, object):
        """ Returns whether this object has children (True for this class).
        """
        return True

    #-------------------------------------------------------------------------
    #  Gets the object's children:
    #-------------------------------------------------------------------------

    def get_children(self, object):
        """ Gets the object's children.
        """
        return [(object, node) for node in self.nodes]

    #-------------------------------------------------------------------------
    #  Gets the object's children identifier:
    #-------------------------------------------------------------------------

    def get_children_id(self, object):
        """ Gets the object's children identifier.
        """
        return ''

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children replaced' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_replaced(self, object, listener, remove):
        """ Sets up or removes a listener for children being replaced on a
        specified object.
        """
        pass

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'children changed' on a specified
    #  object:
    #-------------------------------------------------------------------------

    def when_children_changed(self, object, listener, remove):
        """ Sets up or removes a listener for children being changed on a
        specified object.
        """
        pass

    #-------------------------------------------------------------------------
    #  Gets the label to display for a specified object:
    #-------------------------------------------------------------------------

    def get_label(self, object):
        """ Gets the label to display for a specified object.
        """
        return self.root_node.get_label(object)

    #-------------------------------------------------------------------------
    #  Sets the label for a specified object:
    #-------------------------------------------------------------------------

    def set_label(self, object, label):
        """ Sets the label for a specified object.
        """
        return self.root_node.set_label(object, label)

    #-------------------------------------------------------------------------
    #  Sets up/Tears down a listener for 'label changed' on a specified object:
    #-------------------------------------------------------------------------

    def when_label_changed(self, object, listener, remove):
        """ Sets up or removes a listener for the label being changed on a
        specified object.
        """
        return self.root_node.when_label_changed(object, listener, remove)

    #-------------------------------------------------------------------------
    #  Returns the icon for a specified object:
    #-------------------------------------------------------------------------

    def get_icon(self, object, is_expanded):
        """ Returns the icon for a specified object.
        """
        return self.root_node.get_icon(object, is_expanded)

    #-------------------------------------------------------------------------
    #  Returns the path used to locate an object's icon:
    #-------------------------------------------------------------------------

    def get_icon_path(self, object):
        """ Returns the path used to locate an object's icon.
        """
        return self.root_node.get_icon_path(object)

    #-------------------------------------------------------------------------
    #  Returns the name to use when adding a new object instance (displayed in
    #  the 'New' submenu):
    #-------------------------------------------------------------------------

    def get_name(self, object):
        """ Returns the name to use when adding a new object instance
            (displayed in the "New" submenu).
        """
        return self.root_node.get_name(object)

    #-------------------------------------------------------------------------
    #  Gets the View to use when editing an object:
    #-------------------------------------------------------------------------

    def get_view(self, object):
        """ Gets the view to use when editing an object.
        """
        return self.root_node.get_view(object)

    #-------------------------------------------------------------------------
    #  Returns the right-click context menu for an object:
    #-------------------------------------------------------------------------

    def get_menu(self, object):
        """ Returns the right-click context menu for an object.
        """
        return self.root_node.get_menu(object)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be renamed:
    #-------------------------------------------------------------------------

    def can_rename(self, object):
        """ Returns whether the object's children can be renamed (False for
        this class).
        """
        return False

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be renamed:
    #-------------------------------------------------------------------------

    def can_rename_me(self, object):
        """ Returns whether the object can be renamed (False for this class).
        """
        return False

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be copied:
    #-------------------------------------------------------------------------

    def can_copy(self, object):
        """ Returns whether the object's children can be copied.
        """
        return self.root_node.can_copy(object)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be deleted:
    #-------------------------------------------------------------------------

    def can_delete(self, object):
        """ Returns whether the object's children can be deleted (False for
        this class).
        """
        return False

    #-------------------------------------------------------------------------
    #  Returns whether or not the object can be deleted:
    #-------------------------------------------------------------------------

    def can_delete_me(self, object):
        """ Returns whether the object can be deleted (True for this class).
        """
        return True

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children can be inserted (or just
    #  appended):
    #-------------------------------------------------------------------------

    def can_insert(self, object):
        """ Returns whether the object's children can be inserted (False,
        meaning that children are appended, for this class).
        """
        return False

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-opened:
    #-------------------------------------------------------------------------

    def can_auto_open(self, object):
        """ Returns whether the object's children should be automatically
        opened.
        """
        return self.root_node.can_auto_open(object)

    #-------------------------------------------------------------------------
    #  Returns whether or not the object's children should be auto-closed:
    #-------------------------------------------------------------------------

    def can_auto_close(self, object):
        """ Returns whether the object's children should be automatically
        closed.
        """
        return self.root_node.can_auto_close(object)

    #-------------------------------------------------------------------------
    #  Returns whether a given 'add_object' can be added to an object:
    #-------------------------------------------------------------------------

    def can_add(self, object, add_object):
        """ Returns whether a given object is droppable on the node (False for
        this class).
        """
        return False

    #-------------------------------------------------------------------------
    #  Returns the list of classes that can be added to the object:
    #-------------------------------------------------------------------------

    def get_add(self, object):
        """ Returns the list of classes that can be added to the object.
        """
        return []

    #-------------------------------------------------------------------------
    #  Returns the 'draggable' version of a specified object:
    #-------------------------------------------------------------------------

    def get_drag_object(self, object):
        """ Returns a draggable version of a specified object.
        """
        return self.root_node.get_drag_object(object)

    #-------------------------------------------------------------------------
    #  Returns a droppable version of a specified object:
    #-------------------------------------------------------------------------

    def drop_object(self, object, dropped_object):
        """ Returns a droppable version of a specified object.
        """
        return self.root_node.drop_object(object, dropped_object)

    #-------------------------------------------------------------------------
    #  Handles an object being selected:
    #-------------------------------------------------------------------------

    def select(self, object):
        """ Handles an object being selected.
        """
        return self.root_node.select(object)

    #-------------------------------------------------------------------------
    #  Handles an object being clicked:
    #-------------------------------------------------------------------------

    def click(self, object):
        """ Handles an object being clicked.
        """
        return self.root_node.click(object)

    #-------------------------------------------------------------------------
    #  Handles an object being double-clicked:
    #-------------------------------------------------------------------------

    def dclick(self, object):
        """ Handles an object being double-clicked.
        """
        return self.root_node.dclick(object)

    #-------------------------------------------------------------------------
    #  Handles an object being activated:
    #-------------------------------------------------------------------------

    def activated(self, object):
        """ Handles an object being activated.
        """
        return self.root_node.activated(object)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2013, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Corran Webster
#  Date:   Oct 2013
#
#------------------------------------------------------------------------------

"""
Test case for bug (wx, Mac OS X)

A ListStrEditor was not checking for valid item indexes under Wx.  This was
most noticeable when the selected_index was set in the editor factory.
"""

from traits.has_traits import HasTraits
from traits.trait_types import List, Int, Str
from traitsui.item import Item
from traitsui.view import View
from traitsui.editors.list_str_editor import ListStrEditor

from traitsui.tests._tools import (
    press_ok_button, skip_if_not_qt4, skip_if_not_wx,
    store_exceptions_on_all_threads)


class ListStrEditorWithSelectedIndex(HasTraits):
    values = List(Str())
    selected_index = Int()
    selected_indices = List(Int())
    selected = Str()

single_select_view = View(
    Item('values',
         show_label=False,
         editor=ListStrEditor(
             selected_index='selected_index',
             editable=False),
         ),
    buttons=['OK'])

multi_select_view = View(
    Item('values',
         show_label=False,
         editor=ListStrEditor(
             multi_select=True,
             selected_index='selected_indices',
             editable=False),
         ),
    buttons=['OK'])

single_select_item_view = View(
    Item('values',
         show_label=False,
         editor=ListStrEditor(
             selected='selected',
             editable=False),
         ),
    buttons=['OK'])


def get_selected(control):
    """ Returns a list of the indices of all currently selected list items.
    """
    import wx
    selected = []
    item = -1
    while True:
        item = control.GetNextItem(item, wx.LIST_NEXT_ALL,
                                   wx.LIST_STATE_SELECTED)
        if item == -1:
            break
        selected.append(item)
    return selected


@skip_if_not_wx
def test_wx_list_str_selected_index():
    # behavior: when starting up, the

    with store_exceptions_on_all_threads():
        obj = ListStrEditorWithSelectedIndex(
            values=['value1', 'value2'],
            selected_index=1)
        ui = obj.edit_traits(view=single_select_view)

        # the following is equivalent to setting the text in the text control,
        # then pressing OK

        liststrctrl = ui.control.FindWindowByName('listCtrl')
        selected_1 = get_selected(liststrctrl)

        obj.selected_index = 0
        selected_2 = get_selected(liststrctrl)

        # press the OK button and close the dialog
        press_ok_button(ui)

    # the number traits should be between 3 and 8
    assert selected_1 == [1]
    assert selected_2 == [0]


@skip_if_not_wx
def test_wx_list_str_multi_selected_index():
    # behavior: when starting up, the

    with store_exceptions_on_all_threads():
        obj = ListStrEditorWithSelectedIndex(
            values=['value1', 'value2'],
            selected_indices=[1])
        ui = obj.edit_traits(view=multi_select_view)

        # the following is equivalent to setting the text in the text control,
        # then pressing OK

        liststrctrl = ui.control.FindWindowByName('listCtrl')
        selected_1 = get_selected(liststrctrl)

        obj.selected_indices = [0]
        selected_2 = get_selected(liststrctrl)

        # press the OK button and close the dialog
        press_ok_button(ui)

    # the number traits should be between 3 and 8
    assert selected_1 == [1]
    assert selected_2 == [0]


@skip_if_not_qt4
def test_selection_listener_disconnected():
    """ Check that selection listeners get correctly disconnected """
    from pyface.api import GUI
    from pyface.qt.QtGui import QApplication, QItemSelectionModel
    from pyface.ui.qt4.util.event_loop_helper import EventLoopHelper
    from pyface.ui.qt4.util.testing import event_loop

    obj = ListStrEditorWithSelectedIndex(values=['value1', 'value2'])

    with store_exceptions_on_all_threads():
        qt_app = QApplication.instance()
        if qt_app is None:
            qt_app = QApplication([])
        helper = EventLoopHelper(gui=GUI(), qt_app=qt_app)

        # open the UI and run until the dialog is closed
        ui = obj.edit_traits(view=single_select_item_view)
        with helper.delete_widget(ui.control):
            press_ok_button(ui)

        # now run again and change the selection
        ui = obj.edit_traits(view=single_select_item_view)
        with event_loop():
            editor = ui.get_editors('values')[0]

            list_view = editor.list_view
            mi = editor.model.index(1)
            list_view.selectionModel().select(mi, QItemSelectionModel.ClearAndSelect)

    obj.selected = 'value2'


if __name__ == '__main__':
    # Executing the file opens the dialog for manual testing
    editor = ListStrEditorWithSelectedIndex(
        values=['value1', 'value2'],
        selected_index=1)
    editor.configure_traits(view=single_select_view)



""" Classes used to represent assignment statements. """


# Standard library imports.
import ast

# Enthought library imports.
from traits.api import Any, Bool, HasTraits, Instance, Int, List, Str

# Local imports.
from envisage._compat import STRING_BASE_CLASS


class Assign(HasTraits):
    """ An assignment statement. """

    #### 'Assign' interface ###################################################

    # The namespace that the assignment statement is in.
    namespace = Instance(
        'envisage.developer.code_browser.namespace.Namespace'
    )

    # The line number within the module at which the assignment statement
    # appears.
    lineno = Int

    # The names being assigned to (in Python there can be more than one).
    targets = List(Str)

    # The expression being assigned to the targets (an AST node).
    expr = Any

    # We only care about assignments to trait types, not literals or
    # expressions etc.
    source = Str

    # Is this a trait assignment?
    is_trait = Bool(False)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns an informal string representation of the object. """

        return 'Assign %s at %d' % (', '.join(self.targets), self.lineno)


class AssignFactory(HasTraits):
    """ A factory for assignment statements. """

    ###########################################################################
    # 'AssignFactory' interface.
    ###########################################################################

    def from_ast(self, namespace, node):
        """ Creates an assignment statement from an AST node. """

        # Create a new assignment statement.
        assign = Assign(
            namespace = namespace,
            lineno    = node.lineno,
            expr      = node.value
        )

        # Walk the AST picking out the things we care about!
        AssignVisitor(assign).visit(node)

        return assign


class AssignVisitor(ast.NodeVisitor):
    """ An AST visitor for assigment statements. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, assign):
        """ Creates a new visitor. """

        self.assign = assign

        return

    ###########################################################################
    # 'NodeVisitor' interface.
    ###########################################################################

    def visit_Assign(self, node):
        """ Visits an assignment node. """
        for name in node.targets:
            self.assign.targets.append(name.id)

        self.assign.source = self._get_name(node.value)

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_name(self, node):
        """ Returns the (possibly dotted) name from a node. """

        # fixme: Work out when node can be none here!!!!! I think it is safe
        # to ignore it in terms of working out what is a trait, but it would
        # be nice to know what is going on ;^)
        if node is not None:
            if isinstance(node, STRING_BASE_CLASS):
                name = node

            elif not isinstance(node, ast.AST):
                name = ''

            elif isinstance(node, ast.Name):
                name = node.id

            else:
                names = [self._get_name(child) for child in ast.walk(node) if child is not node]
                name = '.'.join(names)

        else:
            name = ''

        return name

#### EOF ######################################################################



""" (Pyface) Action for displaying a help example.

    :Copyright: 2008, Enthought Inc.
    :License: BSD
    :Author: Janet Swisher
"""
# This software is provided without warranty under the terms of the BSD
# license included in AppTools/trunk/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!


# Standard library imports.
import logging
from subprocess import Popen

# Enthought library imports.
from envisage.api import IExtensionPointUser, IExtensionRegistry
from pyface.workbench.action.workbench_action import WorkbenchAction
from traits.api import Instance, provides, Property

from apptools.help.help_plugin.api import HelpCode
from apptools.help.help_plugin.examples_preferences import \
    ExamplesPreferences

# Local import
from util import get_sys_prefix_relative_filename

# Logging.
logger = logging.getLogger(__name__)

# This module's parent package.
PARENT = '.'.join(__name__.split('.')[:-2])

@provides(IExtensionPointUser)
class ExampleAction(WorkbenchAction):
    """ (Pyface) Action for displaying a help example.
    """


    ### IExtensionPointUser interface
    extension_registry = Property(Instance(IExtensionRegistry))

    def _get_extension_registry(self):
        return self.window.application.extension_registry

    ### HelpExampleAction interface

    # Help code associated with this action.
    my_help_code = Instance(HelpCode)

    # Preferences for examples
    preferences = Instance(ExamplesPreferences)

    def _my_help_code_default(self):
        exns = self.extension_registry.get_extensions(PARENT + '.help_examples')
        for he in exns:
            if he.label == self.name:
                return he
        return None

    def perform(self, event):
        """ Perform the action by displaying the document.
        """
        if self.my_help_code is not None:
            filename = get_sys_prefix_relative_filename(self.my_help_code.filename)
            if filename is not None:
                logger.info('Perform ExampleAction on %s' % filename)
                if self.preferences.editor is not None:
                    # Run the editor, passing it the filename
                    try:
                        Popen([self.preferences.editor, filename])
                    except OSError, err:
                        logger.error(
                        'Could not execute program for Example "%s":\n\n ' \
                        % self.my_help_code.label + str(err) + '\n\nTry ' +\
                        'changing Examples Preferences.')



#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt

import nose
import numpy as np
from numpy.testing import assert_array_equal

try:
    from pandas import DataFrame
except ImportError as exc:
    print "Can't import Pandas: skipping"
    raise nose.SkipTest

from traits.api import Event, HasTraits, Instance

from traitsui.item import Item
from traitsui.ui_editors.data_frame_editor import (
    DataFrameEditor, DataFrameAdapter)
from traitsui.view import View

from traitsui.tests._tools import store_exceptions_on_all_threads, skip_if_null


class DataFrameViewer(HasTraits):

    data = Instance('pandas.core.frame.DataFrame')

    view = View(
        Item('data', editor=DataFrameEditor(), width=400)
    )


format_mapping_view = View(
    Item('data', editor=DataFrameEditor(formats={'X': '%05d', 'Y': '%s'}),
         width=400)
)


font_mapping_view = View(
    Item('data', editor=DataFrameEditor(fonts={'X': 'Courier 10 bold',
                                               'Y': 'Swiss'}),
         width=400)
)


columns_view = View(
    Item('data', editor=DataFrameEditor(columns=['X', ('Zed', 'Z'), 'missing']),  # noqa
            width=400)
)

DATA = [[ 0,  1,  2],
        [ 3,  4,  5],
        [ 6,  7,  8],
        [ 9, 10, 11]]


def sample_data():
    df = DataFrame(DATA, index=['one', 'two', 'three', 'four'],
                   columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=df)
    return viewer


def sample_data_numerical_index():
    df = DataFrame(DATA, index=range(1,5), columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=df)
    return viewer


def sample_text_data():
    data = [[0, 1, 'two'],
            [3, 4, 'five'],
            [6, 7, 'eight'],
            [9, 10, 'eleven']]
    df = DataFrame(data, index=['one', 'two', 'three', 'four'],
                   columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=df)
    return viewer


@skip_if_null
def test_adapter_get_item():
    viewer = sample_data()
    adapter = DataFrameAdapter()

    item_0_df = adapter.get_item(viewer, 'data', 0)

    assert_array_equal(item_0_df.values, [[0, 1, 2]])
    assert_array_equal(item_0_df.columns, ['X', 'Y', 'Z'])
    assert item_0_df.index[0] == 'one'


@skip_if_null
def test_adapter_empty_dataframe():
    data = DataFrame()
    viewer = DataFrameViewer(data=data)
    adapter = DataFrameAdapter()

    item_0_df = adapter.get_item(viewer, 'data', 0)

    assert_array_equal(item_0_df.values, np.array([]).reshape(0, 0))
    assert_array_equal(item_0_df.columns, [])


@skip_if_null
def test_adapter_no_rows():
    data = DataFrame(columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=data)
    adapter = DataFrameAdapter()

    item_0_df = adapter.get_item(viewer, 'data', 0)

    assert_array_equal(item_0_df.values, np.array([]).reshape(0, 3))
    assert_array_equal(item_0_df.columns, ['X', 'Y', 'Z'])


@skip_if_null
def test_adapter_get_item_numerical():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()

    item_0_df = adapter.get_item(viewer, 'data', 0)

    assert_array_equal(item_0_df.values, [[0, 1, 2]])
    assert_array_equal(item_0_df.columns, ['X', 'Y', 'Z'])
    assert item_0_df.index[0] == 1


@skip_if_null
def test_adapter_delete_start():
    viewer = sample_data()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 0)
    data = viewer.data

    assert_array_equal(data.values,
                       [[3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['two', 'three', 'four'])


@skip_if_null
def test_adapter_delete_start_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 0)
    data = viewer.data

    assert_array_equal(data.values,
                       [[3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [2, 3, 4])


@skip_if_null
def test_adapter_delete_middle():
    viewer = sample_data()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 1)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['one', 'three', 'four'])


@skip_if_null
def test_adapter_delete_middle_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 1)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [1, 3, 4])


@skip_if_null
def test_adapter_delete_end():
    viewer = sample_data()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 3)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['one', 'two', 'three'])


@skip_if_null
def test_adapter_delete_end_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()

    adapter.delete(viewer, 'data', 3)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [1, 2, 3])


@skip_if_null
def test_adapter_insert_start():
    viewer = sample_data()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=['new'], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 0, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[-3, -2, -1],
                        [0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['new', 'one', 'two', 'three', 'four'])


@skip_if_null
def test_adapter_insert_start_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=[0], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 0, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[-3, -2, -1],
                        [0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [0, 1, 2, 3, 4])


@skip_if_null
def test_adapter_insert_middle():
    viewer = sample_data()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=['new'], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 1, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [-3, -2, -1],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['one', 'new', 'two', 'three', 'four'])


@skip_if_null
def test_adapter_insert_middle_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=[0], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 1, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [-3, -2, -1],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [1, 0, 2, 3, 4])


@skip_if_null
def test_adapter_insert_end():
    viewer = sample_data()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=['new'], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 5, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11],
                        [-3, -2, -1]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, ['one', 'two', 'three', 'four', 'new'])


@skip_if_null
def test_adapter_insert_end_numerical_index():
    viewer = sample_data_numerical_index()
    adapter = DataFrameAdapter()
    item = DataFrame([[-3, -2, -1]], index=[0], columns=['X', 'Y', 'Z'])

    adapter.insert(viewer, 'data', 5, item)
    data = viewer.data

    assert_array_equal(data.values,
                       [[0, 1, 2],
                        [3, 4, 5],
                        [6, 7, 8],
                        [9, 10, 11],
                        [-3, -2, -1]])
    assert_array_equal(data.columns, ['X', 'Y', 'Z'])
    assert_array_equal(data.index, [1, 2, 3, 4, 0])


@skip_if_null
def test_data_frame_editor():
    viewer = sample_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits()
        ui.dispose()


@skip_if_null
def test_data_frame_editor_alternate_adapter():
    class AlternateAdapter(DataFrameAdapter):
        pass

    alternate_adapter_view = View(
        Item('data', editor=DataFrameEditor(adapter=AlternateAdapter()),
                width=400)
    )
    viewer = sample_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits(view=alternate_adapter_view)
        ui.dispose()


@skip_if_null
def test_data_frame_editor_numerical_index():
    viewer = sample_data_numerical_index()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits()
        ui.dispose()


@skip_if_null
def test_data_frame_editor_text_data():
    viewer = sample_text_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits()
        ui.dispose()


@skip_if_null
def test_data_frame_editor_format_mapping():
    viewer = sample_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits(view=format_mapping_view)
        ui.dispose()


@skip_if_null
def test_data_frame_editor_font_mapping():
    viewer = sample_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits(view=font_mapping_view)
        ui.dispose()


@skip_if_null
def test_data_frame_editor_columns():
    viewer = sample_data()
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits(view=columns_view)
        ui.dispose()


@skip_if_null
def test_data_frame_editor_with_update_refresh():

    class DataFrameViewer(HasTraits):
        data = Instance(DataFrame)
        df_updated = Event
        view = View(
            Item('data', editor=DataFrameEditor(update="df_updated"))
        )

    df = DataFrame(DATA, index=['one', 'two', 'three', 'four'],
                   columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=df)
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits()
        viewer.df_updated = True
        ui.dispose()


@skip_if_null
def test_data_frame_editor_with_refresh():
    class DataFrameViewer(HasTraits):
        data = Instance(DataFrame)
        df_updated = Event
        view = View(
            Item('data', editor=DataFrameEditor())
        )

    df = DataFrame(DATA, index=['one', 'two', 'three', 'four'],
                   columns=['X', 'Y', 'Z'])
    viewer = DataFrameViewer(data=df)
    with store_exceptions_on_all_threads():
        ui = viewer.edit_traits()
        viewer.df_updated = True
        ui.dispose()



#: Number of points used to define a bezier curve
BEZIER_POINTS = 12

#: Number of points used to define a circle
CIRCLE_POINTS = 48

#: Distance under which neighboring points may merge
TOLERANCE = 0.1

#: Default fill color (as per svg spec)
DEFAULT_FILL = [0, 0, 0, 255]

#: Default fill stroke (as per svg spec)
DEFAULT_STROKE = None

#: Size of the texture to render patterns into
PATTERN_TEX_SIZE = 1024

#: svg namespace
XMLNS = 'http://www.w3.org/2000/svg'

#: All possible color names in SVG as hex
named_colors = {'aliceblue': '#F0F8FF',
                'antiquewhite': '#FAEBD7',
                'aqua': '#00FFFF',
                'aquamarine': '#7FFFD4',
                'azure': '#F0FFFF',
                'beige': '#F5F5DC',
                'bisque': '#FFE4C4',
                'black': '#000000',
                'blanchedalmond': '#FFEBCD',
                'blue': '#0000FF',
                'blueviolet': '#8A2BE2',
                'brown': '#A52A2A',
                'burlywood': '#DEB887',
                'cadetblue': '#5F9EA0',
                'chartreuse': '#7FFF00',
                'chocolate': '#D2691E',
                'coral': '#FF7F50',
                'cornflowerblue': '#6495ED',
                'cornsilk': '#FFF8DC',
                'crimson': '#DC143C',
                'cyan': '#00FFFF',
                'darkblue': '#00008B',
                'darkcyan': '#008B8B',
                'darkgoldenrod': '#B8860B',
                'darkgray': '#A9A9A9',
                'darkgreen': '#006400',
                'darkgrey': '#A9A9A9',
                'darkkhaki': '#BDB76B',
                'darkmagenta': '#8B008B',
                'darkolivegreen': '#556B2F',
                'darkorange': '#FF8C00',
                'darkorchid': '#9932CC',
                'darkred': '#8B0000',
                'darksalmon': '#E9967A',
                'darkseagreen': '#8FBC8F',
                'darkslateblue': '#483D8B',
                'darkslategray': '#2F4F4F',
                'darkslategrey': '#2F4F4F',
                'darkturquoise': '#00CED1',
                'darkviolet': '#9400D3',
                'deeppink': '#FF1493',
                'deepskyblue': '#00BFFF',
                'dimgray': '#696969',
                'dimgrey': '#696969',
                'dodgerblue': '#1E90FF',
                'firebrick': '#B22222',
                'floralwhite': '#FFFAF0',
                'forestgreen': '#228B22',
                'fuchsia': '#FF00FF',
                'gainsboro': '#DCDCDC',
                'ghostwhite': '#F8F8FF',
                'gold': '#FFD700',
                'goldenrod': '#DAA520',
                'gray': '#808080',
                'green': '#008000',
                'greenyellow': '#ADFF2F',
                'grey': '#808080',
                'honeydew': '#F0FFF0',
                'hotpink': '#FF69B4',
                'indianred': '#CD5C5C',
                'indigo': '#4B0082',
                'ivory': '#FFFFF0',
                'khaki': '#F0E68C',
                'lavender': '#E6E6FA',
                'lavenderblush': '#FFF0F5',
                'lawngreen': '#7CFC00',
                'lemonchiffon': '#FFFACD',
                'lightblue': '#ADD8E6',
                'lightcoral': '#F08080',
                'lightcyan': '#E0FFFF',
                'lightgoldenrodyellow': '#FAFAD2',
                'lightgray': '#D3D3D3',
                'lightgreen': '#90EE90',
                'lightgrey': '#D3D3D3',
                'lightpink': '#FFB6C1',
                'lightsalmon': '#FFA07A',
                'lightseagreen': '#20B2AA',
                'lightskyblue': '#87CEFA',
                'lightslategray': '#778899',
                'lightslategrey': '#778899',
                'lightsteelblue': '#B0C4DE',
                'lightyellow': '#FFFFE0',
                'lime': '#00FF00',
                'limegreen': '#32CD32',
                'linen': '#FAF0E6',
                'magenta': '#FF00FF',
                'maroon': '#800000',
                'mediumaquamarine': '#66CDAA',
                'mediumblue': '#0000CD',
                'mediumorchid': '#BA55D3',
                'mediumpurple': '#9370DB',
                'mediumseagreen': '#3CB371',
                'mediumslateblue': '#7B68EE',
                'mediumspringgreen': '#00FA9A',
                'mediumturquoise': '#48D1CC',
                'mediumvioletred': '#C71585',
                'midnightblue': '#191970',
                'mintcream': '#F5FFFA',
                'mistyrose': '#FFE4E1',
                'moccasin': '#FFE4B5',
                'navajowhite': '#FFDEAD',
                'navy': '#000080',
                'oldlace': '#FDF5E6',
                'olive': '#808000',
                'olivedrab': '#6B8E23',
                'orange': '#FFA500',
                'orangered': '#FF4500',
                'orchid': '#DA70D6',
                'palegoldenrod': '#EEE8AA',
                'palegreen': '#98FB98',
                'paleturquoise': '#AFEEEE',
                'palevioletred': '#DB7093',
                'papayawhip': '#FFEFD5',
                'peachpuff': '#FFDAB9',
                'peru': '#CD853F',
                'pink': '#FFC0CB',
                'plum': '#DDA0DD',
                'powderblue': '#B0E0E6',
                'purple': '#800080',
                'red': '#FF0000',
                'rosybrown': '#BC8F8F',
                'royalblue': '#4169E1',
                'saddlebrown': '#8B4513',
                'salmon': '#FA8072',
                'sandybrown': '#F4A460',
                'seagreen': '#2E8B57',
                'seashell': '#FFF5EE',
                'sienna': '#A0522D',
                'silver': '#C0C0C0',
                'skyblue': '#87CEEB',
                'slateblue': '#6A5ACD',
                'slategray': '#708090',
                'slategrey': '#708090',
                'snow': '#FFFAFA',
                'springgreen': '#00FF7F',
                'steelblue': '#4682B4',
                'tan': '#D2B48C',
                'teal': '#008080',
                'thistle': '#D8BFD8',
                'tomato': '#FF6347',
                'turquoise': '#40E0D0',
                'violet': '#EE82EE',
                'wheat': '#F5DEB3',
                'white': '#FFFFFF',
                'whitesmoke': '#F5F5F5',
                'yellow': '#FFFF00',
                'yellowgreen': '#9ACD32'}


#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Feb 2012
#
#------------------------------------------------------------------------------

"""
Test the layout when element appear and disappear with visible_when.
"""

from traits.has_traits import HasTraits
from traits.trait_types import Enum, Bool, Str

from traitsui.group import HGroup, VGroup
from traitsui.include import Include
from traitsui.item import Item
from traitsui.view import View

from traitsui.tests._tools import *

_TEXT_WIDTH = 200
_TEXT_HEIGHT = 100


class VisibleWhenProblem(HasTraits):

    which = Enum('one', 'two')

    on = Bool
    txt = Str

    onoff_group = HGroup(
        VGroup(
            Item('on', resizable=False, width=-100, height=-70),
            show_left=False,
            show_border=True, visible_when='which == "one"'
        ),
    )

    text_group = VGroup(
        Item('txt', width=-_TEXT_WIDTH, height=-_TEXT_HEIGHT),
        visible_when='which == "two"',
        show_border=True,
    )

    traits_view = View(
        Item('which'),
        VGroup(
            Include('onoff_group'),
            Include('text_group'),
        ),
        resizable=True,
        buttons=['OK', 'Cancel']
    )


@skip_if_null
def test_visible_when_layout():
    # Bug: The size of a dialog that contains elements that are activated
    # by "visible_when" can end up being the *sum* of the sizes of the
    # elements, even though the elements are mutually exclusive (e.g.,
    # a typical case is a dropbox that lets you select different cases).
    # The expected behavior is that the size of the dialog should be at most
    # the size of the largest combination of elements.

    with store_exceptions_on_all_threads():
        dialog = VisibleWhenProblem()
        ui = dialog.edit_traits()

        # have the dialog switch from group one to two and back to one
        dialog.which = 'two'
        dialog.which = 'one'

        # the size of the window should not be larger than the largest
        # combination (in this case, the `text_group` plus the `which` item
        size = get_dialog_size(ui.control)
        # leave some margin for labels, dropbox, etc
        nose.tools.assert_less(size[0], _TEXT_WIDTH + 100)
        nose.tools.assert_less(size[1], _TEXT_HEIGHT + 150)


if __name__ == '__main__':
    # Execute from command line for manual testing
    vw = VisibleWhenProblem(txt='ciao')
    ui = vw.configure_traits()



"""

  fs.tests.test_archivefs:  testcases for the ArchiveFS class

"""

import unittest
import os
import random
import zipfile
import tempfile
import shutil

import fs.tests
from fs.path import *
try:
    from fs.contrib import archivefs
except ImportError:
    libarchive_available = False
else:
    libarchive_available = True


from six import PY3, b

class TestReadArchiveFS(unittest.TestCase):

    __test__ = libarchive_available

    def setUp(self):
        self.temp_filename = "".join(random.choice("abcdefghijklmnopqrstuvwxyz") for _ in range(6))+".zip"
        self.temp_filename = os.path.join(tempfile.gettempdir(), self.temp_filename)

        self.zf = zipfile.ZipFile(self.temp_filename, "w")
        zf = self.zf
        zf.writestr("a.txt", b("Hello, World!"))
        zf.writestr("b.txt", b("b"))
        zf.writestr("1.txt", b("1"))
        zf.writestr("foo/bar/baz.txt", b("baz"))
        zf.writestr("foo/second.txt", b("hai"))
        zf.close()
        self.fs = archivefs.ArchiveFS(self.temp_filename, "r")

    def tearDown(self):
        self.fs.close()
        os.remove(self.temp_filename)

    def check(self, p):
        try:
            self.zipfile.getinfo(p)
            return True
        except:
            return False

    def test_reads(self):
        def read_contents(path):
            f = self.fs.open(path)
            contents = f.read()
            return contents
        def check_contents(path, expected):
            self.assert_(read_contents(path)==expected)
        check_contents("a.txt", b("Hello, World!"))
        check_contents("1.txt", b("1"))
        check_contents("foo/bar/baz.txt", b("baz"))

    def test_getcontents(self):
        def read_contents(path):
            return self.fs.getcontents(path)
        def check_contents(path, expected):
            self.assert_(read_contents(path)==expected)
        check_contents("a.txt", b("Hello, World!"))
        check_contents("1.txt", b("1"))
        check_contents("foo/bar/baz.txt", b("baz"))

    def test_is(self):
        self.assert_(self.fs.isfile('a.txt'))
        self.assert_(self.fs.isfile('1.txt'))
        self.assert_(self.fs.isfile('foo/bar/baz.txt'))
        self.assert_(self.fs.isdir('foo'))
        self.assert_(self.fs.isdir('foo/bar'))
        self.assert_(self.fs.exists('a.txt'))
        self.assert_(self.fs.exists('1.txt'))
        self.assert_(self.fs.exists('foo/bar/baz.txt'))
        self.assert_(self.fs.exists('foo'))
        self.assert_(self.fs.exists('foo/bar'))

    def test_listdir(self):
        def check_listing(path, expected):
            dir_list = self.fs.listdir(path)
            self.assert_(sorted(dir_list) == sorted(expected))
            for item in dir_list:
                self.assert_(isinstance(item,unicode))
        check_listing('/', ['a.txt', '1.txt', 'foo', 'b.txt'])
        check_listing('foo', ['second.txt', 'bar'])
        check_listing('foo/bar', ['baz.txt'])


class TestWriteArchiveFS(unittest.TestCase):

    __test__ = libarchive_available

    def setUp(self):
        self.temp_filename = "".join(random.choice("abcdefghijklmnopqrstuvwxyz") for _ in range(6))+".zip"
        self.temp_filename = os.path.join(tempfile.gettempdir(), self.temp_filename)

        archive_fs = archivefs.ArchiveFS(self.temp_filename, format='zip', mode='w')

        def makefile(filename, contents):
            if dirname(filename):
                archive_fs.makedir(dirname(filename), recursive=True, allow_recreate=True)
            f = archive_fs.open(filename, 'wb')
            f.write(contents)
            f.close()

        makefile("a.txt", b("Hello, World!"))
        makefile("b.txt", b("b"))
        makefile(u"\N{GREEK SMALL LETTER ALPHA}/\N{GREEK CAPITAL LETTER OMEGA}.txt", b("this is the alpha and the omega"))
        makefile("foo/bar/baz.txt", b("baz"))
        makefile("foo/second.txt", b("hai"))

        archive_fs.close()

    def tearDown(self):
        os.remove(self.temp_filename)

    def test_valid(self):
        zf = zipfile.ZipFile(self.temp_filename, "r")
        self.assert_(zf.testzip() is None)
        zf.close()

    def test_creation(self):
        zf = zipfile.ZipFile(self.temp_filename, "r")
        def check_contents(filename, contents):
            if PY3:
                zcontents = zf.read(filename)
            else:
                zcontents = zf.read(filename.encode(archivefs.ENCODING))
            self.assertEqual(contents, zcontents)
        check_contents("a.txt", b("Hello, World!"))
        check_contents("b.txt", b("b"))
        check_contents("foo/bar/baz.txt", b("baz"))
        check_contents("foo/second.txt", b("hai"))
        check_contents(u"\N{GREEK SMALL LETTER ALPHA}/\N{GREEK CAPITAL LETTER OMEGA}.txt", b("this is the alpha and the omega"))


#~ class TestAppendArchiveFS(TestWriteArchiveFS):

    #~ __test__ = libarchive_available

    #~ def setUp(self):
        #~ self.temp_filename = "".join(random.choice("abcdefghijklmnopqrstuvwxyz") for _ in range(6))+".zip"
        #~ self.temp_filename = os.path.join(tempfile.gettempdir(), self.temp_filename)

        #~ zip_fs = zipfs.ZipFS(self.temp_filename, 'w')

        #~ def makefile(filename, contents):
            #~ if dirname(filename):
                #~ zip_fs.makedir(dirname(filename), recursive=True, allow_recreate=True)
            #~ f = zip_fs.open(filename, 'wb')
            #~ f.write(contents)
            #~ f.close()

        #~ makefile("a.txt", b("Hello, World!"))
        #~ makefile("b.txt", b("b"))

        #~ zip_fs.close()
        #~ zip_fs = zipfs.ZipFS(self.temp_filename, 'a')

        #~ makefile("foo/bar/baz.txt", b("baz"))
        #~ makefile(u"\N{GREEK SMALL LETTER ALPHA}/\N{GREEK CAPITAL LETTER OMEGA}.txt", b("this is the alpha and the omega"))
        #~ makefile("foo/second.txt", b("hai"))

        #~ zip_fs.close()

#~ class TestArchiveFSErrors(unittest.TestCase):

    #~ __test__ = libarchive_available

    #~ def setUp(self):
        #~ self.workdir = tempfile.mkdtemp()

    #~ def tearDown(self):
        #~ shutil.rmtree(self.workdir)

    #~ def test_bogus_zipfile(self):
        #~ badzip = os.path.join(self.workdir,"bad.zip")
        #~ f = open(badzip,"wb")
        #~ f.write(b("I'm not really a zipfile"))
        #~ f.close()
        #~ self.assertRaises(zipfs.ZipOpenError,zipfs.ZipFS,badzip)

    #~ def test_missing_zipfile(self):
        #~ missingzip = os.path.join(self.workdir,"missing.zip")
        #~ self.assertRaises(zipfs.ZipNotFoundError,zipfs.ZipFS,missingzip)


if __name__ == '__main__':
    unittest.main()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2012, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Author: Pietro Berkes
#  Date:   Jan 2012
#
#------------------------------------------------------------------------------

from __future__ import print_function

from traits.has_traits import HasTraits
from traits.trait_types import Float, List, Instance
from traitsui.handler import ModelView
from traitsui.view import View
from traitsui.item import Item
from traitsui.editors.csv_list_editor import CSVListEditor
import traitsui.editors.csv_list_editor as csv_list_editor

from traitsui.tests._tools import *


class ListOfFloats(HasTraits):
    data = List(Float)


class ListOfFloatsWithCSVEditor(ModelView):
    model = Instance(ListOfFloats)

    traits_view = View(
        Item(label="Close the window to append data"),
        Item('model.data', editor=CSVListEditor()),
        buttons=['OK']
    )


@skip_if_null
def test_csv_editor_disposal():
    # Bug: CSVListEditor does not un-hook the traits notifications after its
    # disposal, causing errors when the hooked data is accessed after
    # the window is closed (Issue #48)

    try:
        with store_exceptions_on_all_threads():
            list_of_floats = ListOfFloats(data=[1, 2, 3])
            csv_view = ListOfFloatsWithCSVEditor(model=list_of_floats)
            ui = csv_view.edit_traits()
            press_ok_button(ui)

            # raise an exception if still hooked
            list_of_floats.data.append(2)

    except AttributeError:
        # if all went well, we should not be here
        assert False, "AttributeError raised"


@skip_if_null
def test_csv_editor_external_append():
    # Behavior: CSV editor is notified when an element is appended to the
    # list externally

    def _wx_get_text_value(ui):
        txt_ctrl = ui.control.FindWindowByName('text')
        return txt_ctrl.GetValue()

    def _qt_get_text_value(ui):
        from pyface import qt
        txt_ctrl = ui.control.findChild(qt.QtGui.QLineEdit)
        return txt_ctrl.text()

    with store_exceptions_on_all_threads():
        list_of_floats = ListOfFloats(data=[1.0])
        csv_view = ListOfFloatsWithCSVEditor(model=list_of_floats)
        ui = csv_view.edit_traits()

        # add element to list, make sure that editor knows about it
        list_of_floats.data.append(3.14)

        # get current value from editor
        if is_current_backend_wx():
            value_str = _wx_get_text_value(ui)
        elif is_current_backend_qt4():
            value_str = _qt_get_text_value(ui)

        expected = csv_list_editor._format_list_str([1.0, 3.14])
        nose.tools.assert_equal(value_str, expected)

        press_ok_button(ui)


if __name__ == '__main__':
    # Executing the file opens the dialog for manual testing
    list_of_floats = ListOfFloats(data=[1, 2, 3])
    csv_view = ListOfFloatsWithCSVEditor(model=list_of_floats)
    csv_view.configure_traits()

    # this call will raise an AttributeError in commit
    # 4ecb2fa8f0ef385d55a2a4062d821b0415777973
    # This is because the editor does not un-hook the traits notifications
    # after its disposal
    list_of_floats.data.append(2)
    print(list_of_floats.data)



#!/usr/bin/env python
"""
fs.base
=======

This module defines the most basic filesystem abstraction, the FS class.
Instances of FS represent a filesystem containing files and directories
that can be queried and manipulated.  To implement a new kind of filesystem,
start by sublcassing the base FS class.

For more information regarding implementing a working PyFilesystem interface, see :ref:`implementers`.

"""

from __future__ import with_statement

__all__ = ['DummyLock',
           'silence_fserrors',
           'NullFile',
           'synchronize',
           'FS',
           'flags_to_mode',
           'NoDefaultMeta']

import os
import os.path
import shutil
import fnmatch
import datetime
import time
import errno
try:
    import threading
except ImportError:
    import dummy_threading as threading

from fs.path import *
from fs.errors import *
from fs.local_functools import wraps

import six
from six import b


class DummyLock(object):
    """A dummy lock object that doesn't do anything.

    This is used as a placeholder when locking is disabled.  We can't
    directly use the Lock class from the dummy_threading module, since
    it attempts to sanity-check the sequence of acquire/release calls
    in a way that breaks when real threading is available.

    """

    def acquire(self, blocking=1):
        """Acquiring a DummyLock always succeeds."""
        return 1

    def release(self):
        """Releasing a DummyLock always succeeds."""
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def silence_fserrors(f, *args, **kwargs):
    """Perform a function call and return ``None`` if an :class:`fs.errors.FSError` is thrown

    :param f: Function to call
    :param args: Parameters to f
    :param kwargs: Keyword parameters to f

    """
    try:
        return f(*args, **kwargs)
    except FSError:
        return None


class NoDefaultMeta(object):
    """A singleton used to signify that there is no default for getmeta"""
    pass


class NullFile(object):
    """A NullFile is a file object that has no functionality.

    Null files are returned by the :meth:`fs.base.FS.safeopen` method in FS objects when the
    file doesn't exist. This can simplify code by negating the need to check
    if a file exists, or handling exceptions.

    """
    def __init__(self):
        self.closed = False

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.closed = True

    def flush(self):
        pass

    def next(self):
        raise StopIteration

    def readline(self, *args, **kwargs):
        return b("")

    def close(self):
        self.closed = True

    def read(self, size=None):
        return b("")

    def seek(self, *args, **kwargs):
        pass

    def tell(self):
        return 0

    def truncate(self, *args, **kwargs):
        return 0

    def write(self, data):
        pass

    def writelines(self, *args, **kwargs):
        pass


def synchronize(func):
    """Decorator to synchronize a method on self._lock."""
    @wraps(func)
    def acquire_lock(self, *args, **kwargs):
        self._lock.acquire()
        try:
            return func(self, *args, **kwargs)
        finally:
            self._lock.release()
    return acquire_lock


class FS(object):
    """The base class for Filesystem abstraction objects.
    An instance of a class derived from FS is an abstraction on some kind of filesystem, such as the OS filesystem or a zip file.

    """

    _meta = {}

    def __init__(self, thread_synchronize=True):
        """The base class for Filesystem objects.

        :param thread_synconize: If True, a lock object will be created for the object, otherwise a dummy lock will be used.
        :type thread_synchronize: bool

        """

        self.closed = False
        super(FS, self).__init__()
        self.thread_synchronize = thread_synchronize
        if thread_synchronize:
            self._lock = threading.RLock()
        else:
            self._lock = DummyLock()

    def __del__(self):
        if not getattr(self, 'closed', True):
            try:
                self.close()
            except:
                pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def cachehint(self, enabled):
        """Recommends the use of caching. Implementations are free to use or
            ignore this value.

        :param enabled: If True the implementation is permitted to aggressively cache directory
            structure / file information. Caching such information can speed up many operations,
            particularly for network based filesystems. The downside of caching is that
            changes made to directories or files outside of this interface may not be picked up immediately.

        """
        pass
    # Deprecating cache_hint in favour of no underscore version, for consistency
    cache_hint = cachehint

    def close(self):
        """Close the filesystem. This will perform any shutdown related
        operations required. This method will be called automatically when
        the filesystem object is garbage collected, but it is good practice
        to call it explicitly so that any attached resourced are freed when they
        are no longer required.

        """
        self.closed = True

    def __getstate__(self):
        #  Locks can't be pickled, so instead we just indicate the
        #  type of lock that should be there.  None == no lock,
        #  True == a proper lock, False == a dummy lock.
        state = self.__dict__.copy()
        lock = state.get("_lock", None)
        if lock is not None:
            if isinstance(lock, threading._RLock):
                state["_lock"] = True
            else:
                state["_lock"] = False
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        lock = state.get("_lock")
        if lock is not None:
            if lock:
                self._lock = threading.RLock()
            else:
                self._lock = DummyLock()

    def getmeta(self, meta_name, default=NoDefaultMeta):
        """Retrieve a meta value associated with an FS object.

        Meta values are a way for an FS implementation to report potentially
        useful information associated with the file system.

        A meta key is a lower case string with no spaces. Meta keys may also
        be grouped in namespaces in a dotted notation, e.g. 'atomic.namespaces'.
        FS implementations aren't obliged to return any meta values, but the
        following are common:

         * *read_only* True if the file system cannot be modified
         * *thread_safe* True if the implementation is thread safe
         * *network* True if the file system requires network access
         * *unicode_paths* True if the file system supports unicode paths
         * *case_insensitive_paths* True if the file system ignores the case of paths
         * *atomic.makedir* True if making a directory is an atomic operation
         * *atomic.rename* True if rename is an atomic operation, (and not implemented as a copy followed by a delete)
         * *atomic.setcontents* True if the implementation supports setting the contents of a file as an atomic operation (without opening a file)
         * *free_space* The free space (in bytes) available on the file system
         * *total_space* The total space (in bytes) available on the file system
         * *virtual* True if the filesystem defers to other filesystems
         * *invalid_path_chars* A string containing characters that may not be used in paths

        FS implementations may expose non-generic meta data through a self-named namespace. e.g. ``"somefs.some_meta"``

        Since no meta value is guaranteed to exist, it is advisable to always supply a
        default value to ``getmeta``.

        :param meta_name: The name of the meta value to retrieve
        :param default: An option default to return, if the meta value isn't present
        :raises `fs.errors.NoMetaError`: If specified meta value is not present, and there is no default

        """
        if meta_name not in self._meta:
            if default is not NoDefaultMeta:
                return default
            raise NoMetaError(meta_name=meta_name)
        return self._meta[meta_name]

    def hasmeta(self, meta_name):
        """Check that a meta value is supported

        :param meta_name: The name of a meta value to check
        :rtype: bool

        """
        try:
            self.getmeta(meta_name)
        except NoMetaError:
            return False
        return True

    def validatepath(self, path):
        """Validate an fs path, throws an :class:`~fs.errors.InvalidPathError` exception if validation fails.

        A path is invalid if it fails to map to a path on the underlaying filesystem. The default
        implementation checks for the presence of any of the characters in the meta value 'invalid_path_chars',
        but implementations may have other requirements for paths.

        :param path: an fs path to validatepath
        :raises `fs.errors.InvalidPathError`: if `path` does not map on to a valid path on this filesystem

        """
        invalid_chars = self.getmeta('invalid_path_chars', default=None)
        if invalid_chars:
            re_invalid_chars = getattr(self, '_re_invalid_chars', None)
            if re_invalid_chars is None:
                self._re_invalid_chars = re_invalid_chars = re.compile('|'.join(re.escape(c) for c in invalid_chars), re.UNICODE)
            if re_invalid_chars.search(path):
                raise InvalidCharsInPathError(path)

    def isvalidpath(self, path):
        """Check if a path is valid on this filesystem

        :param path: an fs path

        """
        try:
            self.validatepath(path)
        except InvalidPathError:
            return False
        else:
            return True

    def getsyspath(self, path, allow_none=False):
        """Returns the system path (a path recognized by the OS) if one is present.

        If the path does not map to a system path (and `allow_none` is False)
        then a NoSysPathError exception is thrown.  Otherwise, the system
        path will be returned as a unicode string.

        :param path: a path within the filesystem
        :param allow_none: if True, this method will return None when there is no system path,
            rather than raising NoSysPathError
        :type allow_none: bool
        :raises `fs.errors.NoSysPathError`: if the path does not map on to a system path, and allow_none is set to False (default)
        :rtype: unicode

        """
        if not allow_none:
            raise NoSysPathError(path=path)
        return None

    def hassyspath(self, path):
        """Check if the path maps to a system path (a path recognized by the OS).

        :param path: path to check
        :returns: True if `path` maps to a system path
        :rtype: bool

        """
        return self.getsyspath(path, allow_none=True) is not None

    def getpathurl(self, path, allow_none=False):
        """Returns a url that corresponds to the given path, if one exists.

        If the path does not have an equivalent URL form (and allow_none is False)
        then a :class:`~fs.errors.NoPathURLError` exception is thrown. Otherwise the URL will be
        returns as an unicode string.

        :param path: a path within the filesystem
        :param allow_none: if true, this method can return None if there is no
            URL form of the given path
        :type allow_none: bool
        :raises `fs.errors.NoPathURLError`: If no URL form exists, and allow_none is False (the default)
        :rtype: unicode

        """
        if not allow_none:
            raise NoPathURLError(path=path)
        return None

    def haspathurl(self, path):
        """Check if the path has an equivalent URL form

        :param path: path to check
        :returns: True if `path` has a URL form
        :rtype: bool

        """
        return self.getpathurl(path, allow_none=True) is not None

    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        """Open a the given path as a file-like object.

        :param path: a path to file that should be opened
        :type path: string
        :param mode: mode of file to open, identical to the mode string used
            in 'file' and 'open' builtins
        :type mode: string
        :param kwargs: additional (optional) keyword parameters that may
            be required to open the file
        :type kwargs: dict

        :rtype: a file-like object

        :raises `fs.errors.ParentDirectoryMissingError`: if an intermediate directory is missing
        :raises `fs.errors.ResourceInvalidError`: if an intermediate directory is an file
        :raises `fs.errors.ResourceNotFoundError`: if the path is not found

        """
        raise UnsupportedError("open file")

    def safeopen(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        """Like :py:meth:`~fs.base.FS.open`, but returns a
        :py:class:`~fs.base.NullFile` if the file could not be opened.

        A ``NullFile`` is a dummy file which has all the methods of a file-like object,
        but contains no data.

        :param path: a path to file that should be opened
        :type path: string
        :param mode: mode of file to open, identical to the mode string used
            in 'file' and 'open' builtins
        :type mode: string
        :param kwargs: additional (optional) keyword parameters that may
            be required to open the file
        :type kwargs: dict

        :rtype: a file-like object

        """
        try:
            f = self.open(path, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline, line_buffering=line_buffering, **kwargs)
        except ResourceNotFoundError:
            return NullFile()
        return f

    def exists(self, path):
        """Check if a path references a valid resource.

        :param path: A path in the filesystem
        :type path: string

        :rtype: bool

        """
        return self.isfile(path) or self.isdir(path)

    def isdir(self, path):
        """Check if a path references a directory.

        :param path: a path in the filesystem
        :type path: string

        :rtype: bool

        """
        raise UnsupportedError("check for directory")

    def isfile(self, path):
        """Check if a path references a file.

        :param path: a path in the filesystem
        :type path: string

        :rtype: bool

        """
        raise UnsupportedError("check for file")

    def __iter__(self):
        """ Iterates over paths returned by :py:meth:`~fs.base.listdir` method with default params. """
        for f in self.listdir():
            yield f

    def listdir(self,
                path="./",
                wildcard=None,
                full=False,
                absolute=False,
                dirs_only=False,
                files_only=False):
        """Lists the the files and directories under a given path.

        The directory contents are returned as a list of unicode paths.

        :param path: root of the path to list
        :type path: string
        :param wildcard: Only returns paths that match this wildcard
        :type wildcard: string containing a wildcard, or a callable that accepts a path and returns a boolean
        :param full: returns full paths (relative to the root)
        :type full: bool
        :param absolute: returns absolute paths (paths beginning with /)
        :type absolute: bool
        :param dirs_only: if True, only return directories
        :type dirs_only: bool
        :param files_only: if True, only return files
        :type files_only: bool

        :rtype: iterable of paths

        :raises `fs.errors.ParentDirectoryMissingError`: if an intermediate directory is missing
        :raises `fs.errors.ResourceInvalidError`: if the path exists, but is not a directory
        :raises `fs.errors.ResourceNotFoundError`: if the path is not found

        """
        raise UnsupportedError("list directory")

    def listdirinfo(self,
                    path="./",
                    wildcard=None,
                    full=False,
                    absolute=False,
                    dirs_only=False,
                    files_only=False):
        """Retrieves a list of paths and path info under a given path.

        This method behaves like listdir() but instead of just returning
        the name of each item in the directory, it returns a tuple of the
        name and the info dict as returned by getinfo.

        This method may be more efficient than calling
        :py:meth:`~fs.base.FS.getinfo` on each individual item returned by :py:meth:`~fs.base.FS.listdir`, particularly
        for network based filesystems.

        :param path: root of the path to list
        :param wildcard: filter paths that match this wildcard
        :param dirs_only: only retrieve directories
        :type dirs_only: bool
        :param files_only: only retrieve files
        :type files_only: bool

        :raises `fs.errors.ResourceNotFoundError`: If the path is not found
        :raises `fs.errors.ResourceInvalidError`: If the path exists, but is not a directory

        """
        path = normpath(path)

        def getinfo(p):
            try:
                if full or absolute:
                    return self.getinfo(p)
                else:
                    return self.getinfo(pathjoin(path, p))
            except FSError:
                return {}

        return [(p, getinfo(p))
                for p in self.listdir(path,
                                      wildcard=wildcard,
                                      full=full,
                                      absolute=absolute,
                                      dirs_only=dirs_only,
                                      files_only=files_only)]

    def _listdir_helper(self,
                        path,
                        entries,
                        wildcard=None,
                        full=False,
                        absolute=False,
                        dirs_only=False,
                        files_only=False):
        """A helper method called by listdir method that applies filtering.

        Given the path to a directory and a list of the names of entries within
        that directory, this method applies the semantics of the listdir()
        keyword arguments. An appropriately modified and filtered list of
        directory entries is returned.

        """
        path = normpath(path)
        if dirs_only and files_only:
            raise ValueError("dirs_only and files_only can not both be True")

        if wildcard is not None:
            if not callable(wildcard):
                wildcard_re = re.compile(fnmatch.translate(wildcard))
                wildcard = lambda fn: bool(wildcard_re.match(fn))
            entries = [p for p in entries if wildcard(p)]

        if dirs_only:
            isdir = self.isdir
            entries = [p for p in entries if isdir(pathcombine(path, p))]
        elif files_only:
            isfile = self.isfile
            entries = [p for p in entries if isfile(pathcombine(path, p))]

        if full:
            entries = [pathcombine(path, p) for p in entries]
        elif absolute:
            path = abspath(path)
            entries = [(pathcombine(path, p)) for p in entries]

        return entries

    def ilistdir(self,
                 path="./",
                 wildcard=None,
                 full=False,
                 absolute=False,
                 dirs_only=False,
                 files_only=False):
        """Generator yielding the files and directories under a given path.

        This method behaves identically to :py:meth:`fs.base.FS.listdir` but returns an generator
        instead of a list.  Depending on the filesystem this may be more
        efficient than calling :py:meth:`fs.base.FS.listdir` and iterating over the resulting list.

        """
        return iter(self.listdir(path,
                                 wildcard=wildcard,
                                 full=full,
                                 absolute=absolute,
                                 dirs_only=dirs_only,
                                 files_only=files_only))

    def ilistdirinfo(self,
                     path="./",
                     wildcard=None,
                     full=False,
                     absolute=False,
                     dirs_only=False,
                     files_only=False):
        """Generator yielding paths and path info under a given path.

        This method behaves identically to :py:meth:`~fs.base.listdirinfo` but returns an generator
        instead of a list.  Depending on the filesystem this may be more
        efficient than calling :py:meth:`~fs.base.listdirinfo` and iterating over the resulting
        list.

        """
        return iter(self.listdirinfo(path,
                                     wildcard,
                                     full,
                                     absolute,
                                     dirs_only,
                                     files_only))

    def makedir(self, path, recursive=False, allow_recreate=False):
        """Make a directory on the filesystem.

        :param path: path of directory
        :type path: string
        :param recursive: if True, any intermediate directories will also be created
        :type recursive: bool
        :param allow_recreate: if True, re-creating a directory wont be an error
        :type allow_create: bool

        :raises `fs.errors.DestinationExistsError`: if the path is already a directory, and allow_recreate is False
        :raises `fs.errors.ParentDirectoryMissingError`: if a containing directory is missing and recursive is False
        :raises `fs.errors.ResourceInvalidError`: if a path is an existing file
        :raises `fs.errors.ResourceNotFoundError`: if the path is not found

        """
        raise UnsupportedError("make directory")

    def remove(self, path):
        """Remove a file from the filesystem.

        :param path: Path of the resource to remove
        :type path: string

        :raises `fs.errors.ParentDirectoryMissingError`: if an intermediate directory is missing
        :raises `fs.errors.ResourceInvalidError`: if the path is a directory
        :raises `fs.errors.ResourceNotFoundError`: if the path does not exist

        """
        raise UnsupportedError("remove resource")

    def removedir(self, path, recursive=False, force=False):
        """Remove a directory from the filesystem

        :param path: path of the directory to remove
        :type path: string
        :param recursive: if True, empty parent directories will be removed
        :type recursive: bool
        :param force: if True, any directory contents will be removed
        :type force: bool

        :raises `fs.errors.DirectoryNotEmptyError`: if the directory is not empty and force is False
        :raises `fs.errors.ParentDirectoryMissingError`: if an intermediate directory is missing
        :raises `fs.errors.ResourceInvalidError`: if the path is not a directory
        :raises `fs.errors.ResourceNotFoundError`: if the path does not exist

        """
        raise UnsupportedError("remove directory")

    def rename(self, src, dst):
        """Renames a file or directory

        :param src: path to rename
        :type src: string
        :param dst: new name
        :type dst: string

        :raises ParentDirectoryMissingError: if a containing directory is missing
        :raises ResourceInvalidError: if the path or a parent path is not a
            directory or src is a parent of dst or one of src or dst is a dir
            and the other don't
        :raises ResourceNotFoundError: if the src path does not exist

        """
        raise UnsupportedError("rename resource")

    @convert_os_errors
    def settimes(self, path, accessed_time=None, modified_time=None):
        """Set the accessed time and modified time of a file

        :param path: path to a file
        :type path: string
        :param accessed_time: the datetime the file was accessed (defaults to current time)
        :type accessed_time: datetime
        :param modified_time: the datetime the file was modified (defaults to current time)
        :type modified_time: datetime

        """

        with self._lock:
            sys_path = self.getsyspath(path, allow_none=True)
            if sys_path is not None:
                now = datetime.datetime.now()
                if accessed_time is None:
                    accessed_time = now
                if modified_time is None:
                    modified_time = now
                accessed_time = int(time.mktime(accessed_time.timetuple()))
                modified_time = int(time.mktime(modified_time.timetuple()))
                os.utime(sys_path, (accessed_time, modified_time))
                return True
            else:
                raise UnsupportedError("settimes")

    def getinfo(self, path):
        """Returns information for a path as a dictionary. The exact content of
        this dictionary will vary depending on the implementation, but will
        likely include a few common values. The following values will be found
        in info dictionaries for most implementations:

         * "size" - Number of bytes used to store the file or directory
         * "created_time" - A datetime object containing the time the resource was created
         * "accessed_time" - A datetime object containing the time the resource was last accessed
         * "modified_time" - A datetime object containing the time the resource was modified

        :param path: a path to retrieve information for
        :type path: string

        :rtype: dict

        :raises `fs.errors.ParentDirectoryMissingError`: if an intermediate directory is missing
        :raises `fs.errors.ResourceInvalidError`: if the path is not a directory
        :raises `fs.errors.ResourceNotFoundError`: if the path does not exist

        """
        raise UnsupportedError("get resource info")

    def getinfokeys(self, path, *keys):
        """Get specified keys from info dict, as returned from `getinfo`. The returned dictionary may
        not contain all the keys that were asked for, if they aren't available.

        This method allows a filesystem to potentially provide a faster way of retrieving these info values if you
        are only interested in a subset of them.

        :param path: a path to retrieve information for
        :param keys: the info keys you would like to retrieve

        :rtype: dict

        """
        info = self.getinfo(path)
        return dict((k, info[k]) for k in keys if k in info)

    def desc(self, path):
        """Returns short descriptive text regarding a path. Intended mainly as
        a debugging aid.

        :param path: A path to describe
        :rtype: str

        """
        #if not self.exists(path):
        #    return ''
        try:
            sys_path = self.getsyspath(path)
        except NoSysPathError:
            return "No description available"
        return sys_path

    def getcontents(self, path, mode='rb', encoding=None, errors=None, newline=None):
        """Returns the contents of a file as a string.

        :param path: A path of file to read
        :param mode: Mode to open file with (should be 'rb' for binary or 't' for text)
        :param encoding: Encoding to use when reading contents in text mode
        :param errors: Unicode errors parameter if text mode is use
        :param newline: Newlines parameter for text mode decoding
        :rtype: str
        :returns: file contents

        """
        if 'r' not in mode:
            raise ValueError("mode must contain 'r' to be readable")
        f = None
        try:
            f = self.open(path, mode=mode, encoding=encoding, errors=errors, newline=newline)
            contents = f.read()
            return contents
        finally:
            if f is not None:
                f.close()

    def _setcontents(self,
                     path,
                     data,
                     encoding=None,
                     errors=None,
                     chunk_size=1024 * 64,
                     progress_callback=None,
                     finished_callback=None):
        """Does the work of setcontents. Factored out, so that `setcontents_async` can use it"""
        if progress_callback is None:
            progress_callback = lambda bytes_written: None
        if finished_callback is None:
            finished_callback = lambda: None

        if not data:
            progress_callback(0)
            self.createfile(path, wipe=True)
            finished_callback()
            return 0

        bytes_written = 0
        progress_callback(0)

        if hasattr(data, 'read'):
            read = data.read
            chunk = read(chunk_size)
            if isinstance(chunk, six.text_type):
                f = self.open(path, 'wt', encoding=encoding, errors=errors)
            else:
                f = self.open(path, 'wb')
            write = f.write
            try:
                while chunk:
                    write(chunk)
                    bytes_written += len(chunk)
                    progress_callback(bytes_written)
                    chunk = read(chunk_size)
            finally:
                f.close()
        else:
            if isinstance(data, six.text_type):
                with self.open(path, 'wt', encoding=encoding, errors=errors) as f:
                    f.write(data)
                    bytes_written += len(data)
            else:
                with self.open(path, 'wb') as f:
                    f.write(data)
                    bytes_written += len(data)
            progress_callback(bytes_written)

        finished_callback()
        return bytes_written

    def setcontents(self, path, data=b'', encoding=None, errors=None, chunk_size=1024 * 64):
        """A convenience method to create a new file from a string or file-like object

        :param path: a path of the file to create
        :param data: a string or bytes object containing the contents for the new file
        :param encoding: if `data` is a file open in text mode, or a text string, then use this `encoding` to write to the destination file
        :param errors: if `data` is a file open in text mode or a text string, then use `errors` when opening the destination file
        :param chunk_size: Number of bytes to read in a chunk, if the implementation has to resort to a read / copy loop

        """
        return self._setcontents(path, data, encoding=encoding, errors=errors, chunk_size=1024 * 64)

    def setcontents_async(self,
                          path,
                          data,
                          encoding=None,
                          errors=None,
                          chunk_size=1024 * 64,
                          progress_callback=None,
                          finished_callback=None,
                          error_callback=None):
        """Create a new file from a string or file-like object asynchronously

        This method returns a ``threading.Event`` object. Call the ``wait`` method on the event object
        to block until all data has been written, or simply ignore it.

        :param path: a path of the file to create
        :param data: a string or a file-like object containing the contents for the new file
        :param encoding: if `data` is a file open in text mode, or a text string, then use this `encoding` to write to the destination file
        :param errors: if `data` is a file open in text mode or a text string, then use `errors` when opening the destination file
        :param chunk_size: Number of bytes to read and write in a chunk
        :param progress_callback: A function that is called periodically
            with the number of bytes written.
        :param finished_callback: A function that is called when all data has been written
        :param error_callback: A function that is called with an exception
            object if any error occurs during the copy process.
        :returns: An event object that is set when the copy is complete, call
            the `wait` method of this object to block until the data is written

        """

        finished_event = threading.Event()

        def do_setcontents():
            try:
                self._setcontents(path,
                                  data,
                                  encoding=encoding,
                                  errors=errors,
                                  chunk_size=1024 * 64,
                                  progress_callback=progress_callback,
                                  finished_callback=finished_callback)
            except Exception, e:
                if error_callback is not None:
                    error_callback(e)
            finally:
                finished_event.set()

        threading.Thread(target=do_setcontents).start()
        return finished_event

    def createfile(self, path, wipe=False):
        """Creates an empty file if it doesn't exist

        :param path: path to the file to create
        :param wipe: if True, the contents of the file will be erased

        """
        with self._lock:
            if not wipe and self.isfile(path):
                return
            f = None
            try:
                f = self.open(path, 'wb')
            finally:
                if f is not None:
                    f.close()

    def opendir(self, path):
        """Opens a directory and returns a FS object representing its contents.

        :param path: path to directory to open
        :type path: string

        :return: the opened dir
        :rtype: an FS object

        """

        from fs.wrapfs.subfs import SubFS
        if not self.exists(path):
            raise ResourceNotFoundError(path)
        if not self.isdir(path):
            raise ResourceInvalidError("path should reference a directory")
        return SubFS(self, path)

    def walk(self,
             path="/",
             wildcard=None,
             dir_wildcard=None,
             search="breadth",
             ignore_errors=False):
        """Walks a directory tree and yields the root path and contents.
        Yields a tuple of the path of each directory and a list of its file
        contents.

        :param path: root path to start walking
        :type path: string
        :param wildcard: if given, only return files that match this wildcard
        :type wildcard: a string containing a wildcard (e.g. `*.txt`) or a callable that takes the file path and returns a boolean
        :param dir_wildcard: if given, only walk directories that match the wildcard
        :type dir_wildcard: a string containing a wildcard (e.g. `*.txt`) or a callable that takes the directory name and returns a boolean
        :param search: a string identifying the method used to walk the directories. There are two such methods:

             * ``"breadth"`` yields paths in the top directories first
             * ``"depth"`` yields the deepest paths first

        :param ignore_errors: ignore any errors reading the directory
        :type ignore_errors: bool

        :rtype: iterator of (current_path, paths)

        """

        path = normpath(path)

        if not self.exists(path):
            raise ResourceNotFoundError(path)

        def listdir(path, *args, **kwargs):
            if ignore_errors:
                try:
                    return self.listdir(path, *args, **kwargs)
                except:
                    return []
            else:
                return self.listdir(path, *args, **kwargs)

        if wildcard is None:
            wildcard = lambda f: True
        elif not callable(wildcard):
            wildcard_re = re.compile(fnmatch.translate(wildcard))
            wildcard = lambda fn: bool(wildcard_re.match(fn))

        if dir_wildcard is None:
            dir_wildcard = lambda f: True
        elif not callable(dir_wildcard):
            dir_wildcard_re = re.compile(fnmatch.translate(dir_wildcard))
            dir_wildcard = lambda fn: bool(dir_wildcard_re.match(fn))

        if search == "breadth":
            dirs = [path]
            dirs_append = dirs.append
            dirs_pop = dirs.pop
            isdir = self.isdir
            while dirs:
                current_path = dirs_pop()
                paths = []
                paths_append = paths.append
                try:
                    for filename in listdir(current_path, dirs_only=True):
                        path = pathcombine(current_path, filename)
                        if dir_wildcard(path):
                            dirs_append(path)
                    for filename in listdir(current_path, files_only=True):
                        path = pathcombine(current_path, filename)
                        if wildcard(filename):
                            paths_append(filename)
                except ResourceNotFoundError:
                    # Could happen if another thread / process deletes something whilst we are walking
                    pass

                yield (current_path, paths)

        elif search == "depth":

            def recurse(recurse_path):
                try:
                    for path in listdir(recurse_path, wildcard=dir_wildcard, full=True, dirs_only=True):
                        for p in recurse(path):
                            yield p
                except ResourceNotFoundError:
                    # Could happen if another thread / process deletes something whilst we are walking
                    pass
                yield (recurse_path, listdir(recurse_path, wildcard=wildcard, files_only=True))

            for p in recurse(path):
                yield p

        else:
            raise ValueError("Search should be 'breadth' or 'depth'")

    def walkfiles(self,
                  path="/",
                  wildcard=None,
                  dir_wildcard=None,
                  search="breadth",
                  ignore_errors=False):
        """Like the 'walk' method, but just yields file paths.

        :param path: root path to start walking
        :type path: string
        :param wildcard: if given, only return files that match this wildcard
        :type wildcard: A string containing a wildcard (e.g. `*.txt`) or a callable that takes the file path and returns a boolean
        :param dir_wildcard: if given, only walk directories that match the wildcard
        :type dir_wildcard: A string containing a wildcard (e.g. `*.txt`) or a callable that takes the directory name and returns a boolean
        :param search: a string identifying the method used to walk the directories. There are two such methods:

             * ``"breadth"`` yields paths in the top directories first
             * ``"depth"`` yields the deepest paths first

        :param ignore_errors: ignore any errors reading the directory
        :type ignore_errors: bool

        :rtype: iterator of file paths

        """
        for path, files in self.walk(normpath(path), wildcard=wildcard, dir_wildcard=dir_wildcard, search=search, ignore_errors=ignore_errors):
            for f in files:
                yield pathcombine(path, f)

    def walkdirs(self,
                 path="/",
                 wildcard=None,
                 search="breadth",
                 ignore_errors=False):
        """Like the 'walk' method but yields directories.

        :param path: root path to start walking
        :type path: string
        :param wildcard: if given, only return directories that match this wildcard
        :type wildcard: A string containing a wildcard (e.g. `*.txt`) or a callable that takes the directory name and returns a boolean
        :param search: a string identifying the method used to walk the directories. There are two such methods:

             * ``"breadth"`` yields paths in the top directories first
             * ``"depth"`` yields the deepest paths first

        :param ignore_errors: ignore any errors reading the directory
        :type ignore_errors: bool

        :rtype: iterator of dir paths

        """
        for p, _files in self.walk(path, dir_wildcard=wildcard, search=search, ignore_errors=ignore_errors):
            yield p

    def getsize(self, path):
        """Returns the size (in bytes) of a resource.

        :param path: a path to the resource
        :type path: string

        :returns: the size of the file
        :rtype: integer

        """
        info = self.getinfo(path)
        size = info.get('size', None)
        if size is None:
            raise OperationFailedError("get size of resource", path)
        return size

    def copy(self, src, dst, overwrite=False, chunk_size=1024 * 64):
        """Copies a file from src to dst.

        :param src: the source path
        :type src: string
        :param dst: the destination path
        :type dst: string
        :param overwrite: if True, then an existing file at the destination may
            be overwritten; If False then DestinationExistsError
            will be raised.
        :type overwrite: bool
        :param chunk_size: size of chunks to use if a simple copy is required
            (defaults to 64K).
        :type chunk_size: bool

        """
        with self._lock:
            if not self.isfile(src):
                if self.isdir(src):
                    raise ResourceInvalidError(src, msg="Source is not a file: %(path)s")
                raise ResourceNotFoundError(src)
            if not overwrite and self.exists(dst):
                raise DestinationExistsError(dst)

            src_syspath = self.getsyspath(src, allow_none=True)
            dst_syspath = self.getsyspath(dst, allow_none=True)

            if src_syspath is not None and dst_syspath is not None:
                self._shutil_copyfile(src_syspath, dst_syspath)
            else:
                src_file = None
                try:
                    src_file = self.open(src, "rb")
                    self.setcontents(dst, src_file, chunk_size=chunk_size)
                except ResourceNotFoundError:
                    if self.exists(src) and not self.exists(dirname(dst)):
                        raise ParentDirectoryMissingError(dst)
                finally:
                    if src_file is not None:
                        src_file.close()

    @classmethod
    @convert_os_errors
    def _shutil_copyfile(cls, src_syspath, dst_syspath):
        try:
            shutil.copyfile(src_syspath, dst_syspath)
        except IOError, e:
            #  shutil reports ENOENT when a parent directory is missing
            if getattr(e, "errno", None) == errno.ENOENT:
                if not os.path.exists(dirname(dst_syspath)):
                    raise ParentDirectoryMissingError(dst_syspath)
            raise

    @classmethod
    @convert_os_errors
    def _shutil_movefile(cls, src_syspath, dst_syspath):
        shutil.move(src_syspath, dst_syspath)


    def move(self, src, dst, overwrite=False, chunk_size=16384):
        """moves a file from one location to another.

        :param src: source path
        :type src: string
        :param dst: destination path
        :type dst: string
        :param overwrite: When True the destination will be overwritten (if it exists),
            otherwise a DestinationExistsError will be thrown
        :type overwrite: bool
        :param chunk_size: Size of chunks to use when copying, if a simple copy
            is required
        :type chunk_size: integer

        :raise `fs.errors.DestinationExistsError`: if destination exists and `overwrite` is False

        """

        with self._lock:
            src_syspath = self.getsyspath(src, allow_none=True)
            dst_syspath = self.getsyspath(dst, allow_none=True)

            #  Try to do an os-level rename if possible.
            #  Otherwise, fall back to copy-and-remove.
            if src_syspath is not None and dst_syspath is not None:
                if not os.path.isfile(src_syspath):
                    if os.path.isdir(src_syspath):
                        raise ResourceInvalidError(src, msg="Source is not a file: %(path)s")
                    raise ResourceNotFoundError(src)
                if not overwrite and os.path.exists(dst_syspath):
                    raise DestinationExistsError(dst)
                try:
                    os.rename(src_syspath, dst_syspath)
                    return
                except OSError:
                    pass
            self.copy(src, dst, overwrite=overwrite, chunk_size=chunk_size)
            self.remove(src)

    def movedir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        """moves a directory from one location to another.

        :param src: source directory path
        :type src: string
        :param dst: destination directory path
        :type dst: string
        :param overwrite: if True then any existing files in the destination
            directory will be overwritten
        :type overwrite: bool
        :param ignore_errors: if True then this method will ignore FSError
            exceptions when moving files
        :type ignore_errors: bool
        :param chunk_size: size of chunks to use when copying, if a simple copy
            is required
        :type chunk_size: integer

        :raise `fs.errors.DestinationExistsError`: if destination exists and `overwrite` is False

        """
        with self._lock:
            if not self.isdir(src):
                if self.isfile(src):
                    raise ResourceInvalidError(src, msg="Source is not a directory: %(path)s")
                raise ResourceNotFoundError(src)
            if not overwrite and self.exists(dst):
                raise DestinationExistsError(dst)

            src_syspath = self.getsyspath(src, allow_none=True)
            dst_syspath = self.getsyspath(dst, allow_none=True)

            if src_syspath is not None and dst_syspath is not None:
                try:
                    os.rename(src_syspath, dst_syspath)
                    return
                except OSError:
                    pass

            def movefile_noerrors(src, dst, **kwargs):
                try:
                    return self.move(src, dst, **kwargs)
                except FSError:
                    return
            if ignore_errors:
                movefile = movefile_noerrors
            else:
                movefile = self.move

            src = abspath(src)
            dst = abspath(dst)

            if dst:
                self.makedir(dst, allow_recreate=overwrite)

            for dirname, filenames in self.walk(src, search="depth"):

                dst_dirname = relpath(frombase(src, abspath(dirname)))
                dst_dirpath = pathjoin(dst, dst_dirname)
                self.makedir(dst_dirpath, allow_recreate=True, recursive=True)

                for filename in filenames:

                    src_filename = pathjoin(dirname, filename)
                    dst_filename = pathjoin(dst_dirpath, filename)
                    movefile(src_filename, dst_filename, overwrite=overwrite, chunk_size=chunk_size)

                self.removedir(dirname)

    def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        """copies a directory from one location to another.

        :param src: source directory path
        :type src: string
        :param dst: destination directory path
        :type dst: string
        :param overwrite: if True then any existing files in the destination
            directory will be overwritten
        :type overwrite: bool
        :param ignore_errors: if True, exceptions when copying will be ignored
        :type ignore_errors: bool
        :param chunk_size: size of chunks to use when copying, if a simple copy
            is required (defaults to 16K)

        """
        with self._lock:
            if not self.isdir(src):
                raise ResourceInvalidError(src, msg="Source is not a directory: %(path)s")

            def copyfile_noerrors(src, dst, **kwargs):
                try:
                    return self.copy(src, dst, **kwargs)
                except FSError:
                    return
            if ignore_errors:
                copyfile = copyfile_noerrors
            else:
                copyfile = self.copy

            src = abspath(src)
            dst = abspath(dst)

            if not overwrite and self.exists(dst):
                raise DestinationExistsError(dst)

            if dst:
                self.makedir(dst, allow_recreate=True)

            for dirname, filenames in self.walk(src):
                dst_dirname = relpath(frombase(src, abspath(dirname)))
                dst_dirpath = pathjoin(dst, dst_dirname)
                self.makedir(dst_dirpath, allow_recreate=True, recursive=True)
                for filename in filenames:
                    src_filename = pathjoin(dirname, filename)
                    dst_filename = pathjoin(dst_dirpath, filename)
                    copyfile(src_filename, dst_filename, overwrite=overwrite, chunk_size=chunk_size)

    def isdirempty(self, path):
        """Check if a directory is empty (contains no files or sub-directories)

        :param path: a directory path

        :rtype: bool

        """
        with self._lock:
            path = normpath(path)
            iter_dir = iter(self.ilistdir(path))
            try:
                next(iter_dir)
            except StopIteration:
                return True
            return False

    def makeopendir(self, path, recursive=False):
        """makes a directory (if it doesn't exist) and returns an FS object for
        the newly created directory.

        :param path: path to the new directory
        :param recursive: if True any intermediate directories will be created

        :return: the opened dir
        :rtype: an FS object

        """
        with self._lock:
            self.makedir(path, allow_recreate=True, recursive=recursive)
            dir_fs = self.opendir(path)
            return dir_fs

    def printtree(self, max_levels=5):
        """Prints a tree structure of the FS object to the console

        :param max_levels: The maximum sub-directories to display, defaults to
            5. Set to None for no limit

        """
        from fs.utils import print_fs
        print_fs(self, max_levels=max_levels)
    tree = printtree

    def browse(self, hide_dotfiles=False):
        """Displays the FS tree in a graphical window (requires wxPython)

        :param hide_dotfiles: If True, files and folders that begin with a dot will be hidden

        """
        from fs.browsewin import browse
        browse(self, hide_dotfiles)

    def getmmap(self, path, read_only=False, copy=False):
        """Returns a mmap object for this path.

        See http://docs.python.org/library/mmap.html for more details on the mmap module.

        :param path: A path on this filesystem
        :param read_only: If True, the mmap may not be modified
        :param copy: If False then changes wont be written back to the file
        :raises `fs.errors.NoMMapError`: Only paths that have a syspath can be opened as a mmap

        """
        syspath = self.getsyspath(path, allow_none=True)
        if syspath is None:
            raise NoMMapError(path)

        try:
            import mmap
        except ImportError:
            raise NoMMapError(msg="mmap not supported")

        if read_only:
            f = open(syspath, 'rb')
            access = mmap.ACCESS_READ
        else:
            if copy:
                f = open(syspath, 'rb')
                access = mmap.ACCESS_COPY
            else:
                f = open(syspath, 'r+b')
                access = mmap.ACCESS_WRITE

        m = mmap.mmap(f.fileno(), 0, access=access)
        return m


def flags_to_mode(flags, binary=True):
    """Convert an os.O_* flag bitmask into an FS mode string."""
    if flags & os.O_WRONLY:
        if flags & os.O_TRUNC:
            mode = "w"
        elif flags & os.O_APPEND:
            mode = "a"
        else:
            mode = "r+"
    elif flags & os.O_RDWR:
        if flags & os.O_TRUNC:
            mode = "w+"
        elif flags & os.O_APPEND:
            mode = "a+"
        else:
            mode = "r+"
    else:
        mode = "r"
    if flags & os.O_EXCL:
        mode += "x"
    if binary:
        mode += 'b'
    else:
        mode += 't'
    return mode


